#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Perfect Lens"""

import time
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

start_time = time.time()

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0.0
P_Obj.Thickness = 50
P_Obj.Glass = "AIR"
P_Obj.Diameter = 30.0

# ______________________________________#

L1a = Kos.surf()
L1a.Thin_Lens = 100.
L1a.Thickness = (100 + 50)
L1a.Rc = 0.0
L1a.Glass = "AIR"
L1a.Diameter = 30.0

# ______________________________________#

L1b = Kos.surf()
L1b.Thin_Lens = 50.
L1b.Thickness = 100.
L1b.Rc = 0.0
L1b.Glass = "AIR"
L1b.Diameter = 30.0

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0.0
P_Ima.Thickness = 0.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 100.0
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, L1a, L1b, P_Ima]
config_1 = Kos.Setup()

# ______________________________________#

Doblete = Kos.system(A, config_1)
Rayos1 = Kos.raykeeper(Doblete)
Rayos2 = Kos.raykeeper(Doblete)
Rayos3 = Kos.raykeeper(Doblete)
RayosT = Kos.raykeeper(Doblete)

# ______________________________________#

tam = 10
rad = 10.0
tsis = len(A) - 1
for j in range(-tam, tam + 1):
    for i in range(-tam, tam + 1):
        x_0 = (i / tam) * rad
        y_0 = (j / tam) * rad
        r = np.sqrt((x_0 * x_0) + (y_0 * y_0))
        if r < rad:
            tet = 0.0
            pSource_0 = [x_0, y_0, 0.0]
            dCos = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
            W = 0.4
            Doblete.Trace(pSource_0, dCos, W)
            Rayos1.push()
            RayosT.push()

# ______________________________________#

Kos.display3d(Doblete, RayosT, 0)
X, Y, Z, L, M, N = Rayos1.pick(-1)

# ______________________________________#

plt.plot(X, Y, 'x')
plt.xlabel('X')
plt.ylabel('Y')
plt.title('Stop Diagram')
plt.axis('square')
plt.show()

# ______________________________________#

print("--- %s seconds ---" % (time.time() - start_time))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Aug 15 14:22:55 2021

@author: joelherreravazquez
"""

from AstroAtmosphere import *

# Parameters at Cerro Armazones
T   = 279.65    # k
P   = 71200     # Pa
H   = 0.22
xc  = 450       # ppm
lat = -24.5983  # degrees
h   = 3064      # m
l1  = 1.49      # micron
l2  = 1.78      # micron
z0  = 30

# Initializing dispersion model
at  = Observatory()

# Calculating indices of refraction for l1 and l2
n1  = at.n_tph(l=l1, T=T, p=P, RH=H, xc=xc)
n2  = at.n_tph(l=l2, T=T, p=P, RH=H, xc=xc)

# Density of the atmosphere (following CIPM-81/91 equations)
rho = at.rho(p=P, T=T, RH=H, xc=xc)

# Initializing refraction model and setting the reduced height
disp = dispersion(lat, h)
disp.setReducedHeight(P, rho)

# Calculation of the atmopheric dipsersion
atm_dispersion = disp.cassini(n1, n2, z0) * 3.6e6
print ('The dispersion is %.03f milli arc seconds' %(atm_dispersion))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Doublet Lens Pupil"""

import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# _________________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0.0
P_Obj.Thickness = 100
P_Obj.Glass = "AIR"
P_Obj.Diameter = 30.0
P_Obj.Name = "P_Obj"

# _________________________________________#

L1a = Kos.surf()
L1a.Rc = 9.284706570002484E+001
L1a.Thickness = 6.0
L1a.Glass = "BK7"
L1a.Diameter = 30.0
L1a.Axicon = 0

# _________________________________________#

L1b = Kos.surf()
L1b.Rc = -3.071608670000159E+001
L1b.Thickness = 3.0
L1b.Glass = "F2"
L1b.Diameter = 30

# _________________________________________#

L1c = Kos.surf()
L1c.Rc = -7.819730726078505E+001
L1c.Thickness = 9.737604742910693E+001 - 40
L1c.Glass = "AIR"
L1c.Diameter = 30

# _________________________________________#

pupila = Kos.surf()
pupila.Rc = 0
pupila.Thickness = 40.
pupila.Glass = "AIR"
pupila.Diameter = 3
pupila.Name = "Pupil"
pupila.DespY = 0.
pupila.Nm_Poss=[-10,10]

# _________________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0.0
P_Ima.Thickness = 0.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 20.0
P_Ima.Name = "P_Ima"
P_Ima.Nm_Poss=[-10,10]

# _________________________________________#

A = [P_Obj, L1a, L1b, L1c, pupila, P_Ima]
config_1 = Kos.Setup()

# _________________________________________#

Doblete = Kos.system(A, config_1)
Rayos = Kos.raykeeper(Doblete)

# _________________________________________#

W = 0.4
sup = 4
AperVal = 3
AperType = "EPD"
Pup = Kos.PupilCalc(Doblete, sup, W, AperType, AperVal)

# _________________________________________#

print("Radio pupila de entrada: ")
print(Pup.RadPupInp)
print("Posición pupila de entrada: ")
print(Pup.PosPupInp)
print("Rádio pupila de salida: ")
print(Pup.RadPupOut)
print("Posicion pupila de salida: ")
print(Pup.PosPupOut)
print("Posicion pupila de salida respecto al plano focal: ")
print(Pup.PosPupOutFoc)
print("Orientación pupila de salida")
print(Pup.DirPupSal)

print("Airy disk diameter focal distance (micrometers)")
print(Pup.FocusAiryRadius)

# _________________________________________#

[L, M, N] = Pup.DirPupSal
print(L, M, N)

# _________________________________________#

Pup.Samp = 3
Pup.Ptype = "fan"
Pup.FieldType = "angle"
Pup.FieldY = 2.0
x, y, z, L, M, N = Pup.Pattern2Field()
for i in range(0, len(x)):
    pSource_0 = [x[i], y[i], z[i]]
    dCos = [L[i], M[i], N[i]]
    Doblete.Trace(pSource_0, dCos, W)
    Rayos.push()

# _________________________________________#

Pup.FieldY = -2.0
x, y, z, L, M, N = Pup.Pattern2Field()
for i in range(0, len(x)):
    pSource_0 = [x[i], y[i], z[i]]
    dCos = [L[i], M[i], N[i]]
    Doblete.Trace(pSource_0, dCos, W)
    Rayos.push()

# _________________________________________#

# Kos.display3d(Doblete, Rayos,2)
Kos.display2d(Doblete, Rayos,0,1)



import inspect
from .SystemTools import *
RUTE=inspect.getmodule(load_Catalog).__file__
rute=RUTE[:-15]+ "/Cat/"





class Setup():

    def __init__(self):
        """Kraken_setup.
        """
        print("Catalog dir: ", rute)
        cat1 = (rute + 'SCHOTT.AGF')
        cat2 = (rute + 'infrared.agf')

        self.GlassCat =[]
        self.GlassCat.append(cat1)
        self.GlassCat.append(cat2)

        self.Load(self.GlassCat)

    def Load(self, GL):

        print("Glass catalog loaded")


        utilities = (rute + 'UTILIDADES.AGF')
        GL.append(utilities)


        [self.CAT, self.NAMES, self.NM, self.ED, self.CD, self.TD, self.OD, self.LD, self.IT] = load_Catalog(GL)
        file = (rute + 'Alum.csv')
        (self.W_alum, self.N_alum, self.K_alum) = load_alluminum_complex(file)







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 23 18:07:46 2021

@author: joelherreravazquez
"""
import numpy as np
from scipy.optimize import fsolve


def f(z1):
    r=574
    L=0
    M=0
    N=1
    x0=10.0
    y0=0

    x1 = -z1*(L/N) - x0
    y1 = -z1*(M/N) - y0

    s= np.sqrt(x1**2 +y1**2)
    c= 1/r
    ze = (c*s**2)/np.sqrt(1-((c**2)*(s**2))) - r

    zt = np.abs(ze - z1)

    return(zt)

V=0
root = fsolve(f, V)

print(" Valor con fsolve", root)




z = 0
h= 0.0000001
for i in range(0,5):
    z = z - f(z)/( (f(z+h) - f(z-h) )/ (h*2.))
print("Valor con Newthon - Raphson: ",z)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Tel 2M"""

import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos


import numpy as np
import matplotlib.pyplot as plt
import time

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0
P_Obj.Thickness = 1000 + 3.452200000000000E+003
P_Obj.Glass = "AIR"
P_Obj.Diameter = 1.059E+003 * 2.0

# ______________________________________#

Thickness = 3.452200000000000E+003
M1 = Kos.surf()
M1.Rc = -9.638000000004009E+003
M1.Thickness = -Thickness
M1.k = -1.077310000000000E+000
M1.Glass = "MIRROR"
M1.Diameter = 1.059E+003 * 2.0
M1.InDiameter = 250 * 2.0

# ______________________________________#

M2 = Kos.surf()
M2.Rc = -3.93E+003
M2.Thickness = Thickness + 1.037525880125084E+003
M2.k = -4.328100000000000E+000
M2.Glass = "MIRROR"
M2.Diameter = 3.365E+002 * 2.0
M2.AxisMove = 0

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Diameter = 1000.0
P_Ima.Glass = "AIR"
P_Ima.Name = "Plano imagen"
A = [P_Obj, M1, M2, P_Ima]

# ______________________________________#

configuracion_1 = Kos.Setup()
Telescopio = Kos.system(A, configuracion_1)
Rayos = Kos.raykeeper(Telescopio)

# ______________________________________#

W = 0.4
sup = 1
AperVal = 2010
AperType = "EPD"
Pup = Kos.PupilCalc(Telescopio, sup, W, AperType, AperVal)
Pup.Samp = 7
Pup.FieldType = "angle"

# ______________________________________#

Pup.FieldX = 0.0
Pup.Ptype = "hexapolar"
xa, ya, za, La, Ma, Na = Pup.Pattern2Field()

# ______________________________________#

Pup.FieldX = 0.5
Pup.FieldY = 0.5
Pup.Ptype = "square"
xb, yb, zb, Lb, Mb, Nb = Pup.Pattern2Field()

# ______________________________________#

Pup.FieldX = -.5
Pup.FieldY = -.5
Pup.Ptype = "rand"
xc, yc, zc, Lc, Mc, Nc = Pup.Pattern2Field()

# ______________________________________#

for i in range(0, len(xa)):
    pSource_0 = [xa[i], ya[i], za[i]]
    dCos = [La[i], Ma[i], Na[i]]
    Telescopio.Trace(pSource_0, dCos, W)
    Rayos.push()
# ______________________________________#

for i in range(0, len(xb)):
    pSource_0 = [xb[i], yb[i], zb[i]]
    dCos = [Lb[i], Mb[i], Nb[i]]
    Telescopio.Trace(pSource_0, dCos, W)
    Rayos.push()

# ______________________________________#

for i in range(0, len(xc)):
    pSource_0 = [xc[i], yc[i], zc[i]]
    dCos = [Lc[i], Mc[i], Nc[i]]
    Telescopio.Trace(pSource_0, dCos, W)
    Rayos.push()

# ______________________________________#

Kos.display3d(Telescopio, Rayos, 2)
X, Y, Z, L, M, N = Rayos.pick(-1)

# ______________________________________#

plt.plot(X, Y, 'x')
plt.xlabel('x')
plt.ylabel('y')
plt.title('Spot Diagram')
plt.axis('square')
plt.show()




import numpy as np
import pyvista as pv

class Prerequisites():
    """Prerequisites.
    """


# self.SubAperture[0]


    def __init__(self, SurfData, SUTO):
        """__init__.

        Parameters
        ----------
        SurfData :
            SurfData
        SUTO :
            SUTO
        """
        self.SDT = SurfData
        self.n = len(self.SDT)
        self.ANG = np.arange(0, 362, 1)
        self.ANG_Rad = np.deg2rad(self.ANG)
        self.Disable_Inner = 1
        self.ExtraDiameter = 0
        self.SuTo = SUTO
        self.TRANS_1A = []
        self.TRANS_2A = []

    def GeometricRotatAndTran(self, L_te_h, j):
        """GeometricRotatAndTran.

        Parameters
        ----------
        L_te_h :
            L_te_h
        j :
            j
        """
        stx = 0
        sty = 0
        for n in range(j, (- 1), (- 1)):
            LL = 1.0
            if (n == 0):
                LL = 0
            if (n == j):
                PA = 1
            else:
                PA = (self.SDT[n].AxisMove * LL)
            tx = ((self.SDT[n].TiltX * PA) * LL)
            ty = ((self.SDT[n].TiltY * PA) * LL)
            tz = ((self.SDT[n].TiltZ * PA) * LL)
            dx = ((self.SDT[n].DespX * PA) * LL)
            dy = ((self.SDT[n].DespY * PA) * LL)
            dz = ((self.SDT[n].DespZ * PA) * LL)
            stx = (stx + tx)
            sty = (sty + ty)
            Tol_Err = 0.0001
            if (np.abs(np.cos(np.deg2rad(stx))) < Tol_Err):
                tx = (tx + Tol_Err)
            if (np.abs(np.cos(np.deg2rad(sty))) < Tol_Err):
                ty = (ty + Tol_Err)
            if (self.SDT[n].Order == 0):
                L_te_h.rotate_x(tx, inplace=True)
                L_te_h.rotate_y(ty, inplace=True)
                L_te_h.rotate_z(-tz, inplace=True)
                L_te_h.translate([dx, dy, dz],inplace=True)
                L_te_h.translate([0, 0, self.SDT[(n - 1)].Thickness],inplace=True)
            else:
                L_te_h.translate([dx, dy, dz],inplace=True)
                L_te_h.rotate_z(-tz,inplace=True)
                L_te_h.rotate_y(ty,inplace=True)
                L_te_h.rotate_x(tx,inplace=True)
                L_te_h.translate([0, 0, self.SDT[(n - 1)].Thickness],inplace=True)
        return L_te_h

    def Flat2SigmaSurface(self, plane_object, j):
        """Flat2SigmaSurface.

        Parameters
        ----------
        plane_object :
            plane_object
        j :
            j
        """
        plane_objectAx = plane_object.points[:, 0]
        plane_objectAy = plane_object.points[:, 1]
        plane_objectAz = plane_object.points[:, 2]

        plane_objectAx = np.asarray(plane_objectAx)
        plane_objectAy = np.asarray(plane_objectAy)
        plane_objectAz = np.asarray(plane_objectAz)



        plane_objectAx = plane_objectAx + self.SDT[j].SubAperture[2]
        plane_objectAy = plane_objectAy + self.SDT[j].SubAperture[1]

        plane_objectAz = self.SuTo.SurfaceShape(plane_objectAx, plane_objectAy, j)
        plane_objectC = np.c_[(plane_objectAx, plane_objectAy, plane_objectAz)]
        plane_object.points = plane_objectC
        return plane_object

    def Face3D(self, j):
        """Face3D.

        Parameters
        ----------
        j :
            j
        """
        if self.SDT[j].Mask_Shape == "None":
            self.SDT[j].RestoreVTK()


        if (self.SDT[j].Solid_3d_stl == 'None'):
            RES = (47 * self.SDT[j].Res)
            con = (((self.SDT[j].Diameter *self.SDT[j].SubAperture[0]) - (self.SDT[j].InDiameter * self.Disable_Inner)) / (self.SDT[j].Diameter *self.SDT[j].SubAperture[0]))
            if (con == 0):
                con = 1
            r_RES = int((RES * con))
            INNER = ((self.SDT[j].InDiameter * self.Disable_Inner) / 2.0)
            OUTER = ((self.SDT[j].Diameter *self.SDT[j].SubAperture[0]) / 2.0)
            disc = pv.Disc(center=[0.0, 0.0, 0.0], inner=INNER, outer=OUTER, normal=(0, 0, 1), r_res=r_RES, c_res=(RES * 2))
            L_te_h = self.Flat2SigmaSurface(disc, j)
            if (self.SDT[j].InDiameter > 0):
                L_te_h = L_te_h.delaunay_2d().edge_source = L_te_h
            else:
                L_te_h = L_te_h.delaunay_2d()
        else:
            L_te_h = pv.read(self.SDT[j].Solid_3d_stl)
            L_te_h.compute_normals(cell_normals=True, point_normals=True, split_vertices=True, flip_normals=False, consistent_normals=True, auto_orient_normals=False, non_manifold_traversal=True, feature_angle=30.0, inplace = False)
        L_te_h = self.GeometricRotatAndTran(L_te_h, j)
        MASK = self.SDT[j].Mask_Shape
        OBJECT_MASK = pv.MultiBlock()
        for mask in MASK:
            Mask_poly = self.Flat2SigmaSurface(mask, j)
            Mask_poly = self.GeometricRotatAndTran(Mask_poly, j)
            OBJECT_MASK.append(Mask_poly)
        return (L_te_h, OBJECT_MASK)

    def SidePerim(self, j):
        """SidePerim.

        Parameters
        ----------
        j :
            j
        """
        rad2 = ((self.SDT[j].Diameter *self.SDT[j].SubAperture[0]) / 2.0)

        x2 = rad2 * np.cos(self.ANG_Rad)
        y2 = rad2 * np.sin(self.ANG_Rad)
        z2 = np.zeros_like(x2)
        x2 = x2 + self.SDT[j].SubAperture[2]
        y2 = y2 + self.SDT[j].SubAperture[1]

        points2 = np.c_[(x2, y2, z2)]
        L_te = pv.PolyData(points2, force_float=False)
        x2 = L_te.points[:, 0]
        y2 = L_te.points[:, 1]
        z2 = self.SuTo.SurfaceShape(x2, y2, j)
        points2 = np.c_[(x2, y2, z2)]
        L_te = pv.PolyData(points2, force_float=False)
        L_te = self.GeometricRotatAndTran(L_te, j)
        return L_te.points

    def Side3D(self, j, j2):
        """Side3D.

        Parameters
        ----------
        j :
            j
        j2 :
            j2
        """
        PTS1 = self.SidePerim(j)
        x1 = PTS1[:, 0]
        y1 = PTS1[:, 1]
        z1 = PTS1[:, 2]
        PTS2 = self.SidePerim(j2)
        x2 = PTS2[:, 0]
        y2 = PTS2[:, 1]
        z2 = PTS2[:, 2]

        RR = np.zeros_like(x2)
        for s in range(0, len(x2)):
            x2 = np.roll(x2, 1)
            y2 = np.roll(y2, 1)
            z2 = np.roll(z2, 1)

            X = x2 - x1
            Y = y2 - y1
            Z = z2 - z1
            R = np.sqrt(X**2 + Y**2 + Z**2)


            RR[s] = np.mean(R)
        ag = np.argmin(RR)

        x2 = np.roll(x2, ag)
        y2 = np.roll(y2, ag)
        z2 = np.roll(z2, ag)






        P = []
        F = []
        for i in range(0, 360):
            P.append([x1[i], y1[i], z1[i]])
            P.append([x2[i], y2[i], z2[i]])
            P.append([x1[(i + 1)], y1[(i + 1)], z1[(i + 1)]])
            F.append([3, (0 + (i * 6)), (1 + (i * 6)), (2 + (i * 6))])
            P.append([x1[(i + 1)], y1[(i + 1)], z1[(i + 1)]])
            P.append([x2[i], y2[i], z2[i]])
            P.append([x2[(i + 1)], y2[(i + 1)], z2[(i + 1)]])
            F.append([3, (3 + (i * 6)), (4 + (i * 6)), (5 + (i * 6))])
        P = np.asarray(P)
        F = np.asarray(F)
        cant = pv.PolyData(P, F, force_float=False)
        Ax = cant.points[:, 0]
        Ay = cant.points[:, 1]
        Az = cant.points[:, 2]
        Ax = np.asarray(Ax)
        Ay = np.asarray(Ay)
        Az = np.asarray(Az)

        cant.compute_normals(cell_normals=True, point_normals=True, split_vertices=True, flip_normals=False, consistent_normals=True, auto_orient_normals=False, non_manifold_traversal=True, feature_angle=30.0, inplace = False)

        return cant

    def Prerequisites3DSolids(self):
        """Prerequisites3DSolids.
        """
        self.GlassOnSide = []
        self.PreTypeTotal = []
        self.TypeTotal = []
        self.PreGlassOnSide = []
        self.AAA = pv.MultiBlock()
        self.BBB = pv.MultiBlock()
        self.DDD = pv.MultiBlock()
        self.EEE = pv.MultiBlock()
        self.counter = []
        self.side_number = []
        for j in range(0, self.n):
            (lens, masked) = self.Face3D(j)
            self.AAA.append(lens)
            if (self.SDT[j].Solid_3d_stl == 'None'):
                self.TypeTotal.append(0)
            else:
                self.TypeTotal.append(1)
            self.EEE.append(lens)
            self.counter.append(j)
            self.DDD.append(masked.copy(deep=True))
            self.counter.append(j)
            self.GlassOnSide.append(j)
            if (j < (self.n - 1)):
                if (self.SDT[j].Glass != 'NULL'):
                    if (self.SDT[j].Glass != 'AIR'):
                        if (self.SDT[j].Glass != 'MIRROR'):
                            if (self.SDT[j].Solid_3d_stl == 'None'):
                                j2 = (j + 1)
                                while True:
                                    if ((self.SDT[j].Glass == 'NULL') or (self.SDT[j2].Solid_3d_stl != 'None')):
                                        j2 = (j2 + 1)
                                    else:
                                        break
                                    if (j2 == (self.n - 1)):
                                        break
                                side = self.Side3D(j, j2)
                                self.BBB.append(side)
                                self.PreTypeTotal.append(1)
                                self.PreGlassOnSide.append(j)
                                self.side_number.append(j)
        for i in self.BBB:
            self.EEE.append(i)
        for i in self.PreGlassOnSide:
            self.GlassOnSide.append(i)
        for i in self.PreTypeTotal:
            self.TypeTotal.append(i)

    def Prerequisites3SMath(self):
        """Prerequisites3SMath.
        """
        self.TRANS_1A = []
        self.TRANS_2A = []
        for j in range(0, self.n):
            Start_trans1 = np.matrix([[1.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 1.0]])
            for n in range(j, (- 1), (- 1)):
                LL = 1.0
                if (n == 0):
                    LL = 0.0
                if (n == j):
                    PA = 1
                else:
                    PA = (self.SDT[n].AxisMove * LL)
                Tx = (((- np.deg2rad(self.SDT[n].TiltX)) * PA) * LL)
                Ty = (((- np.deg2rad(self.SDT[n].TiltY)) * PA) * LL)
                Tz = (((- np.deg2rad(self.SDT[n].TiltZ)) * PA) * LL)
                TH = self.SDT[(n - 1)].Thickness
                (dx, dy, dz) = (((self.SDT[n].DespX * PA) * LL), ((self.SDT[n].DespY * PA) * LL), ((self.SDT[n].DespZ * PA) * LL))
                DTH_Z = np.matrix([[1.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, (- TH)], [0.0, 0.0, 0.0, 1.0]])
                Rx_1A = np.matrix([[1.0, 0.0, 0.0, 0.0], [0.0, np.cos((- Tx)), np.sin((- Tx)), 0.0], [0.0, (- np.sin((- Tx))), np.cos((- Tx)), 0.0], [0.0, 0.0, 0.0, 1.0]])
                Ry_1A = np.matrix([[np.cos((- Ty)), 0.0, (- np.sin((- Ty))), 0.0], [0, 1.0, 0.0, 0.0], [np.sin((- Ty)), 0.0, np.cos((- Ty)), 0.0], [0.0, 0.0, 0.0, 1.0]])
                Rz_1A = np.matrix([[np.cos((- Tz)), (- np.sin((- Tz))), 0.0, 0.0], [np.sin((- Tz)), np.cos((- Tz)), 0.0, 0.0], [0.0, 0.0, 1.0, 0.0], [0, 0, 0, 1.0]])
                Dxyz_1A = np.matrix([[1.0, 0.0, 0.0, (- dx)], [0.0, 1.0, 0.0, (- dy)], [0.0, 0.0, 1.0, (- dz)], [0.0, 0.0, 0.0, 1.0]])
                if (self.SDT[n].Order == 0):
                    Start_trans1 = (((((Start_trans1 * Rz_1A) * Ry_1A) * Rx_1A) * Dxyz_1A) * DTH_Z)
                else:
                    Start_trans1 = (((((Dxyz_1A * Rx_1A) * Ry_1A) * Rz_1A) * Start_trans1) * DTH_Z)
            self.TRANS_1A.append(Start_trans1)
            self.TRANS_2A.append(np.linalg.inv(Start_trans1))




import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
import scipy


import pkg_resources
""" Looking for if KrakenOS is installed, if not, it assumes that
an folder downloaded from github is run"""

required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("Not installed")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos
######################################################################################


P_Obj = Kos.surf()

P_Obj.Thickness = 1.0
P_Obj.Glass = "AIR"
P_Obj.Diameter = 100
P_Obj.Drawing = 0



Ronchi = Kos.surf()
Ronchi.Diameter = 100
Ronchi.Thickness = 5000
Ronchi.Glass = "AIR"
Sp=0.5
n= np.arange(-10,10,Sp)
AAA = pv.MultiBlock()

for j in n:
    Rulling = pv.Plane(center=[j, 0, 0], direction=[0, 0, 1], i_size=Sp/2, j_size=110, i_resolution=5, j_resolution=5)
    AAA.append(Rulling)
Ronchi.Mask_Shape = AAA
Ronchi.Mask_Type = 2


Mirror = Kos.surf()
Mirror.Rc = -5000
Mirror.k = -0.03
Mirror.Diameter = 2500
Mirror.Thickness = -5000
Mirror.Glass = "MIRROR"



Ronchi2 = Kos.surf()
Ronchi2.Diameter = 100
Ronchi2.Thickness = -5000
Ronchi2.Glass = "AIR"
Ronchi2.Mask_Shape = AAA
Ronchi2.Mask_Type = 2



P_Ima = Kos.surf()
P_Ima.Rc = 0
P_Ima.Thickness = -.0001
P_Ima.Glass = "AIR"
P_Ima.Diameter = 6000.0
P_Ima.Drawing = 1
P_Ima.Name = "Plano imagen"

A = [P_Obj, Ronchi, Mirror, Ronchi2, P_Ima]

configur = Kos.Setup()
Telescope = Kos.system(A, configur)
Rays = Kos.raykeeper(Telescope)

W = 0.633
Sun = Kos.SourceRnd()

def f(x):
    res = 1
    return res


Sun.field = 15.0

Sun.fun = f
Sun.dim = 1.5
Sun.num = 10000
L, M, N, X, Y, Z = Sun.rays()

Xr = np.zeros_like(L)
Yr = np.zeros_like(L)
Nr = np.zeros_like(L)

con = 0
con2 = 0
for i in range(0, Sun.num):
    if con2 == 10:
        print(100. * i / Sun.num)
        con2 = 0

    pSource_0 = [X[i], Y[i], Z[i]]
    dCos = [L[i], M[i], N[i]]
    Telescope.Trace(pSource_0, dCos, W)

    Xr[con] = Telescope.Hit_x[-1]
    Yr[con] = Telescope.Hit_y[-1]
    Nr[con] = Telescope.SURFACE[-1]
    con = con + 1
    con2 = con2 + 1
    # Rays.push()

args = np.argwhere(Nr == len(A)-1)
plt.plot(Xr[args], Yr[args], '.', c="g", markersize=1)

# axis labeling
plt.xlabel('x')
plt.ylabel('y')

# figure name
plt.title('Spot diagram')
plt.axis('square')
plt.show()

#             Rays.push()
# Kos.display3d(Telescope, Rays, 1)



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Tel 2M Wavefront Fitting"""

import os
import sys
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

currentDirectory = os.getcwd()
sys.path.insert(1, currentDirectory + '/library')

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0
P_Obj.Thickness = 1000 + 3.452200000000000E+003
P_Obj.Glass = "AIR"
P_Obj.Diameter = 1.059E+003 * 2.0

# ______________________________________#

Thickness = 3.452200000000000E+003
M1 = Kos.surf()
M1.Rc = -9.638000000004009E+003
M1.Thickness = -Thickness
M1.k = -1.077310000000000E+000
M1.Glass = "MIRROR"
M1.Diameter = 1.059E+003 * 2.0
M1.InDiameter = 250 * 2.0
M1.TiltY = 0.0
M1.TiltX = 0.0

# ______________________________________#

M1.AxisMove = 0
M2 = Kos.surf()
M2.Rc = -3.93E+003
M2.Thickness = Thickness + 1037.525880
M2.k = -4.328100000000000E+000
M2.Glass = "MIRROR"
M2.Diameter = 3.365E+002 * 2.0
M2.TiltY = 0.0
M2.TiltX = 0.0
M2.DespY = 0.0
M2.DespX = 0.0
M2.AxisMove = 0

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Diameter = 300.0
P_Ima.Glass = "AIR"
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, M1, M2, P_Ima]
configuracion_1 = Kos.Setup()
Telescopio = Kos.system(A, configuracion_1)

# ______________________________________#

Surf = 1
W = 0.5016
AperVal = 2000.
AperType = "EPD"
Pupil = Kos.PupilCalc(Telescopio, Surf, W, AperType, AperVal)
Pupil.Samp = 10
Pupil.Ptype = "hexapolar"
Pupil.FieldX = 0.1
Pupil.FieldY = 0.0

Pupil.FieldType = "angle"

X, Y, Z, P2V = Kos.Phase(Pupil)
print("Peak to valley: ", P2V)
NC = 18
A = np.ones(38)

Zcoef, Mat, RMS2Chief, RMS2Centroid, FITTINGERROR = Kos.Zernike_Fitting(X, Y, Z, A)


for i in range(0, NC):
    print("z", i + 1, "  ", "{0:.6f}".format(float(Zcoef[i])), ":", Mat[i])
# ______________________________________#

# print("RMS: ", "{:.4f}".format(float(w_rms)), " Error del ajuste: ", fitt_error)
# z_coeff[0] = 0
# print("RMS to chief: ", np.sqrt(np.sum(z_coeff * z_coeff)))
# z_coeff[1] = 0
# z_coeff[2] = 0
# print("RMS to centroid: ", np.sqrt(np.sum(z_coeff * z_coeff)))

# #______________________________________#

RR = Kos.raykeeper(Telescopio)
x, y, z, L, M, N = Pupil.Pattern2Field()

# ______________________________________#

for i in range(0, len(x)):
    pSource_0 = [x[i], y[i], z[i]]
    dCos = [L[i], M[i], N[i]]
    Telescopio.Trace(pSource_0, dCos, W)
    RR.push()

# ______________________________________#

Kos.display2d(Telescopio, RR, 1,0)
X, Y, Z, L, M, N = RR.pick(-1)

# ______________________________________#

plt.plot(X, Y, 'x')
plt.xlabel('numbers')
plt.ylabel('values')
plt.title('spot Diagram')
plt.axis('square')
plt.show()

ima = Kos.WavefrontData2Image(Zcoef, 400)

Type = "interferogram"
Kos.ZernikeDataImage2Plot(ima, Type)



Surf = 1
W = 0.5016
AperVal = 2000.
AperType = "STOP"
Pupil = Kos.PupilCalc(Telescopio, Surf, W, AperType, AperVal)
Pupil.Samp = 10
Pupil.Ptype = "hexapolar"
Pupil.FieldX = 0.1
Pupil.FieldY = 0.0


AB = Kos.Seidel(Pupil)

print("--------------------------------------")
print(AB.SCW_AN)
print(AB.SCW_NM)
print(AB.SCW_TOTAL)


class Function2Optimize:
    def __init__(self, ABER):
        self.ABER = ABER

    def Fun(self, K):
        System = self.ABER.SYSTEM
        System.SDT[1].k = K[0]
        System.SDT[2].k = K[1]
        System.SetData()

        self.ABER.calculate()

        System.RestoreData()
        Sph = self.ABER.SAC_TOTAL[0]
        Coma = self.ABER.SAC_TOTAL[1]
        return [Sph, Coma]


SeidelFun = Function2Optimize(AB)

Pupil.FieldY = 0.2

from scipy.optimize import fsolve

root = fsolve(SeidelFun.Fun, [0, 0])
print(root)




import random

def prov(pro):
    """prov.

    Parameters
    ----------
    pro :
        pro
    """
    a_list = [0, 1]
    prob = pro
    distribution = [prob, (1.0 - prob)]
    random_number = random.choices(a_list, distribution)
    return random_number
for i in range(0, 100):
    rnum = prov(0.05)
    print(rnum)




# from numba import jit
# from numba.experimental import jitclass
# from numba import int32, float32

# @jit(forceobj=True)
import numpy as np
from scipy.interpolate import griddata




class extra__surf():
    """extra__surf.
    """


    def __init__(self, C):
        """__init__.

        Parameters
        ----------
        C :
            C
        """
        self.COEF = C[1]
        self.user_surface = C[0]

    def calculate(self, x, y):
        """calculate.

        Parameters
        ----------
        x :
            x
        y :
            y
        """
        Z = self.user_surface(x, y, self.COEF)
        return Z


# @jit(forceobj=True)
def even_asphere(x, y, E):
    """even_asphere.

    Parameters
    ----------
    x :
        x
    y :
        y
    E :
        E
    """
    r = np.sqrt(((x * x) + (y * y)))
    Z = 0.0 * np.zeros_like(x)
    for i in range(1, 9):
        if (E[(i - 1)] != 0):
            Z = (Z + (E[(i - 1)] * np.power(r, ((2.0 * i*1.0) * 1.0))))
    return Z




class aspheric__surf():
    """aspheric__surf.
    """
    def __init__(self, E):
        """__init__.

        Parameters
        ----------
        E :
            E
        """
        self.E = E


    def calculate(self, x, y):
        """calculate.

        Parameters
        ----------
        x :
            x
        y :
            y
        """
        Z = even_asphere(x, y, self.E)
        return Z




class conic__surf(object):
    """conic__surf.
    """


    def __init__(self, R_C, KON, C_RXY_RATIO):
        """__init__.

        Parameters
        ----------
        R_C :
            R_C
        KON :
            KON
        C_RXY_RATIO :
            C_RXY_RATIO
        """
        self.R_C = R_C
        self.KON = KON
        self.C_RXY_RATIO = C_RXY_RATIO



    def calculate(self, x, y):
        """calculate.

        Parameters
        ----------
        x :
            x
        y :
            y
        """

        z = CalculateCon(x, y, self.R_C , self.C_RXY_RATIO, self.KON)

        return z

# @jit(nopython=True)
def CalculateCon(x, y, R_C , C_RXY_RATIO, KON):
    """calculate.

    Parameters
    ----------
    x :
        x
    y :
        y
    """
    if (R_C != 0.0):
        c = (1.0 / R_C)
        s = np.sqrt(((x * x) + ((y * C_RXY_RATIO) * (y * C_RXY_RATIO))))
        InRoot = (1 - (((((KON + 1.0) * c) * c) * s) * s))
        z = (((c * s) * s) / (1.0 + np.sqrt(InRoot)))
    else:
        z = 0.0 * np.zeros_like(x)

    return z



class axicon__surf():
    """axicon__surf.
    """


    def __init__(self, C_RXY_RATIO, AXC):
        """__init__.

        Parameters
        ----------
        C_RXY_RATIO :
            C_RXY_RATIO
        AXC :
            AXC
        """
        self.C_RXY_RATIO = C_RXY_RATIO
        self.AXC = AXC

    def calculate(self, x, y):
        """calculate.

        Parameters
        ----------
        x :
            x
        y :
            y
        """
        # if (self.AXC != 0):
        #     s = np.sqrt(((x * x) + ((y * self.C_RXY_RATIO) * (y * self.C_RXY_RATIO))))
        #     z_axicon = (s * np.tan(np.deg2rad(self.AXC)))
        # else:
        #     z_axicon = np.zeros_like(x)

        z_axicon = CalculateAxic( x, y, self.C_RXY_RATIO, self.AXC)
        return z_axicon

# @jit(nopython=True)
def CalculateAxic( x, y, C_RXY_RATIO, AXC):
    """calculate.

    Parameters
    ----------
    x :
        x
    y :
        y
    """
    if (AXC != 0.0):
        s = np.sqrt(((x * x) + ((y * C_RXY_RATIO) * (y * C_RXY_RATIO))))
        z_axicon = (s * np.tan(np.deg2rad(AXC)))
    else:
        z_axicon = 0.0 * np.zeros_like(x)
    return z_axicon






class error_map__surf():
    """error_map__surf.
    """


    def __init__(self, xValues, yValues, zValues, SPACE):
        """__init__.

        Parameters
        ----------
        xValues :
            xValues
        yValues :
            yValues
        zValues :
            zValues
        SPACE :
            SPACE
        """
        TG = SPACE
        LGMX = np.max(yValues)
        LGMN = np.min(yValues)
        LG = (LGMX - LGMN)
        NPG = int((1 + (LG / TG)))
        s = 40
        VXX = np.arange((((- LG) / 2.0) - (s * TG)), ((LG / 2.0) + (s * TG)), TG)
        VYY = np.arange((((- LG) / 2.0) - (s * TG)), ((LG / 2.0) + (s * TG)), TG)
        NPG = VXX.shape[0]
        (grid_x, grid_y) = np.meshgrid(VXX, VYY)
        Z = np.zeros((NPG, NPG))
        err = 0.01
        cont = 1
        for h in range(0, NPG):
            for k in range(0, NPG):
                Ox = grid_x[(h, k)]
                Oy = grid_y[(h, k)]
                ARGWx = np.argwhere(((xValues < (Ox + err)) & (xValues > (Ox - err))))
                ARGWy = np.argwhere(((yValues < (Oy + err)) & (yValues > (Oy - err))))
                if ((ARGWx.shape[0] > 0) and (ARGWy.shape[0] > 0)):
                    IN = np.intersect1d(ARGWx, ARGWy)
                    if (IN.shape[0] > 0):
                        varg = IN[0]
                        Z[(h, k)] = zValues[varg]
                        cont = (cont + 1)
        pnts = np.argwhere((Z != 0.0))
        values = Z[(pnts[:, 0], pnts[:, 1])]
        Vx = VXX[pnts[:, 0]]
        Vy = VYY[pnts[:, 1]]
        points = np.vstack((Vx, Vy)).T
        aXX = np.arange((((- LG) / 2.0) - (s * TG)), ((LG / 2.0) + (s * TG)), (TG / 1.0))
        aYY = np.arange((((- LG) / 2.0) - (s * TG)), ((LG / 2.0) + (s * TG)), (TG / 1.0))
        (grid_x, grid_y) = np.meshgrid(aXX, aYY)
        grid_z0 = griddata(points, values, (grid_x, grid_y), method='nearest')
        Z[0, :] = grid_z0[0, :]
        Z[(NPG - 1), :] = grid_z0[(NPG - 1), :]
        Z[:, 0] = grid_z0[:, 0]
        Z[:, (NPG - 1)] = grid_z0[:, (NPG - 1)]
        pnts = np.argwhere((Z != 0.0))
        Vx = VXX[pnts[:, 0]]
        Vy = VYY[pnts[:, 1]]
        self.values = Z[(pnts[:, 0], pnts[:, 1])]
        self.points = np.vstack((Vx, Vy)).T

    def calculate(self, x, y):
        """calculate.

        Parameters
        ----------
        x :
            x
        y :
            y
        """
        z = griddata(self.points, self.values, (x, y), method='cubic')
        return z







class zernike__surf():
    """zernike__surf.
    """


    def __init__(self, COEF, Z_POL, Z_POW, DMTR):
        """__init__.

        Parameters
        ----------
        COEF :
            COEF
        Z_POL :
            Z_POL
        Z_POW :
            Z_POW
        DMTR :
            DMTR
        """
        self.COEF = COEF
        self.Z_POL = Z_POL
        self.Z_POW = Z_POW
        self.DMTR = DMTR

    def calculate(self, x, y):
        """calculate.

        Parameters
        ----------
        x :
            x
        y :
            y
        """
        # ZSP = np.zeros_like(x)
        # for i in range(0, self.COEF.shape[0]):
        #     if (self.COEF[i] != 0):
        #         p = (np.sqrt(((x * x) + (y * y))) / (self.DMTR / 2.0))
        #         f = np.arctan2(x, y)
        #         ZSP = (ZSP + (self.COEF[i] * zernike_polynomials(i, p, f, self.Z_POL, self.Z_POW)))

        ZSP = CalculateZern(x, y, self.Z_POL, self.Z_POW, self.COEF, self.DMTR)
        return ZSP



# @jit(forceobj=True)
def CalculateZern( x, y, Z_POL, Z_POW, COEF, DMTR):
    """calculate.

    Parameters
    ----------
    x :
        x
    y :
        y
    """
    ZSP = 0.0*np.zeros_like(x)
    for i in range(0, COEF.shape[0]):
        if (COEF[i] != 0):
            p = (np.sqrt(((x * x) + (y * y))) / (DMTR / 2.0))
            f = np.arctan2(x, y)
            ZSP = (ZSP + (COEF[i] * zernike_polynomials(i, p, f, Z_POL, Z_POW)))
    return ZSP





# @jit(forceobj=True)
def zernike_polynomials(term, ro, theta, Zern_pol, z_pow):
    """zernike_polynomials.

    Parameters
    ----------
    term :
        term
    ro :
        ro
    theta :
        theta
    Zern_pol :
        Zern_pol
    z_pow :
        z_pow
    """
    (j, n, m, par, raiz) = Zern_pol[term]
    ct = z_pow[term][0]
    pot = z_pow[term][1]
    NR = 0
    L = len(ct)
    for i in range(0, L):
        NR = ((ct[i] * np.power(ro, pot[i])) + NR)
    if (par == 1):
        S = (raiz * NR)
    if (par == 3):
        S = ((raiz * NR) * np.cos((m * theta)))
    if (par == 2):
        S = ((raiz * NR) * np.sin((m * theta)))
    return S






# @jit(forceobj=True)
def z_parity(num):
    """z_parity.

    Parameters
    ----------
    num :
        num
    """
    nv = (num / 2.0)
    n = (nv - int(nv))
    if (n == 0):
        v = 2
    else:
        v = 3
    return v







# @jit(forceobj=True)
def r_zern(m, n):
    """r_zern.

    Parameters
    ----------
    m :
        m
    n :
        n
    """
    ls = int(((n - m) / 2.0))
    cont = 0
    TCV = []
    pot = []
    a = []
    for s in range(0, (int(ls) + 1)):
        V1 = (np.power((- 1), s) * np.math.factorial((n - s)))
        V2 = ((np.math.factorial(s) * np.math.factorial((((n + m) / 2.0) - s))) * np.math.factorial((((n - m) / 2.0) - s)))
        TC = (V1 / V2)
        potencia = (n - (2.0 * s))
        TCV.append(TC)
        pot.append(potencia)
        cont = (cont + 1)
    a.append(TCV)
    a.append(pot)
    return a







# @jit(forceobj=True)
def zernike_expand(L):
    """zernike_expand.

    Parameters
    ----------
    L :
        L
    """
    cont = 0
    Z = np.array([0, 0, 0, 0, 0])
    n = L
    m = L
    for i in range(0, n):
        if (cont >= L):
            break
        for j in range(0, m):
            v = ((j - i) / 2.0)
            if ((v - int(v)) == 0):
                if ((i > j) or (i == j)):
                    if (j != 0):
                        v = z_parity(cont)
                        Z = np.vstack((Z, [cont, i, j, v, np.sqrt((2.0 * (i + 1.0)))]))
                        if (cont >= L):
                            break
                        cont = (cont + 1)
                        v = z_parity(cont)
                        Z = np.vstack((Z, [cont, i, j, v, np.sqrt((2.0 * (i + 1.0)))]))
                        if (cont >= L):
                            break
                        cont = (cont + 1)
                    if (j == 0):
                        Z = np.vstack((Z, [cont, i, j, 1, np.sqrt((i + 1.0))]))
                        if (cont >= L):
                            break
                        cont = (cont + 1)
    Z = np.delete(Z, 0, 0)
    E = []
    for i in range(0, Z.shape[0]):
        (j, n, m, paR_c, raiz) = Z[i]
        a = r_zern(m, n)
        E.append(a)
    Z = Z[:L]
    return (Z, E)









# @jit(forceobj=True)
def zernike_math_notation(term, Zern_pol, z_pow):
    """zernike_math_notation.

    Parameters
    ----------
    term :
        term
    Zern_pol :
        Zern_pol
    z_pow :
        z_pow
    """
    (j, n, m, par, raiz) = Zern_pol[term]
    ZZZ = ['Piston', 'Tilt x, (about y axis)', 'Tilt y, (about x axis)', 'Power or Focus', 'Astigmatism y, (45deg)', 'Astigmatism x, (0deg)', 'Coma y', 'Coma x', 'Trefoil y', 'Trefoil x', 'Primary Spherical', 'Secondary Astigmatism x', 'Secondary Astigmatism y', 'Tetrafoil x', 'Tetrafoil y', 'Secondary Coma x', 'Secondary Coma y', 'Secondary Trefoil x', 'Secondary Trefoil y', 'Pentafoil x', 'Pentafoil y', 'Secondary Spherical', 'Tertiary Astigmatism y', 'Tertiary Astigmatism x', 'Secondary Tetrafoil y', 'Secondary Tetrafoil x', '14^(1/2) (p^6) * SIN (6A)', '14^(1/2) (p^6) * COS (6A)', 'Tertiary Coma y', 'Tertiary Coma x', 'Tertiary Trefoil y', 'Tertiary Trefoil x']
    ct = z_pow[term][0]
    pot = z_pow[term][1]
    NR = ' '
    L = len(ct)
    for i in range(0, L):
        if (pot[i] == 0):
            NR = ((str(ct[i]) + '+') + NR)
        else:
            NR = ((((str(ct[i]) + 'r^') + str(int(pot[i]))) + '+') + NR)
    i = len(NR)
    NR = NR[:(- 2)]
    if (m == 1):
        mm = ''
    else:
        mm = str(int(m))
    if (par == 1):
        S = (((str(int((0.01 + (raiz * raiz)))) + '^(1/2)(') + NR) + ')')
    if (par == 3):
        S = ((((((str(int((0.01 + (raiz * raiz)))) + '^(1/2)(') + NR) + ')') + 'cos(') + mm) + 'T)')
    if (par == 2):
        S = ((((((str(int((0.01 + (raiz * raiz)))) + '^(1/2)(') + NR) + ')') + 'sin(') + mm) + 'T)')
    if (term < len(ZZZ)):
        x = ZZZ[[term][0]][:]
    else:
        x = ''
    return ((S + '   ') + x)




# @jit(forceobj=True)
def Wavefront_Zernike_Phase(x, y, COEF):
    """Wavefront_Zernike_Phase.

    Parameters
    ----------
    x :
        x
    y :
        y
    COEF :
        COEF
    """
    NC = len(COEF)
    (Zern_pol, z_pow) = zernike_expand(NC)
    tcoef = COEF.shape[0]
    p = np.sqrt(((x * x) + (y * y)))
    f = np.arctan2(x, y)
    ZFP = 0.0
    for i in range(0, tcoef):
        if (COEF[i] != 0):
            ZFP = (ZFP + (COEF[i] * zernike_polynomials(i, p, f, Zern_pol, z_pow)))
    return ZFP






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Parabole Mirror Shift"""

import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Thickness = 1000.0
P_Obj.Diameter = 300
P_Obj.Drawing = 0

# ______________________________________#

M1 = Kos.surf()
M1.Rc = -2000.0
M1.Thickness = M1.Rc / 2
M1.k = -1.0
M1.Glass = "MIRROR"
M1.Diameter = 300
M1.ShiftY = 200

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Glass = "AIR"
P_Ima.Diameter = 1600.0
P_Ima.Drawing = 0
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, M1, P_Ima]
configuracion_1 = Kos.Setup()

# ______________________________________#

Espejo = Kos.system(A, configuracion_1)
Rayos = Kos.raykeeper(Espejo)

# ______________________________________#

tam = 5
rad = 150.0
tsis = len(A) - 1
for i in range(-tam, tam + 1):
    for j in range(-tam, tam + 1):
        x_0 = (i / tam) * rad
        y_0 = (j / tam) * rad
        r = np.sqrt((x_0 * x_0) + (y_0 * y_0))
        if r < rad:
            tet = 0.0
            pSource_0 = [x_0, y_0, 0.0]
            dCos = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
            W = 0.4
            Espejo.Trace(pSource_0, dCos, W)
            Rayos.push()

# ______________________________________#

Kos.display2d(Espejo, Rayos, 0)



#!/usr/bin/env Python3
from .ciddorModel import Observatory
from .refractionModels import refraction
from .dispersionModels import dispersion
from .neoslalib import sla_refro
from .observatories import observatories
from .refractivityModels import Edlen1953
import numpy as np

'''
@ Author: 		Joost van den Born
@ Contact: 		born@astron.nl
@ Description: 	Some functions for quick testing.
'''
def get_conditions(location):
	'''
		Returns the mean conditions at various site. Currently
		supported locations:
		- 'STANDARD'
		- 'CERRO_ARMAZONES'
		- 'CERRO_PARANAL'
		- 'LA_PALMA'
		- 'LA_SILLA'
		- 'LAS_CAMPANAS'
		- 'MAUNA_KEA'

		Parameters
		----------
		location 	: string
			String denoting the desired location

		Returns
		-------
		observatory : dict
			Dictionary containing the atmospheric parameters
	'''
	observatory = observatories[location]
	return observatory


def Filippenko1982(l1, l2, zenith, conditions='STANDARD', T=288.15, p=101325,
							RH=0.0, xc=450, lat=0, h=0, f=None):
	'''
		Calculates the refraction following the 1982 paper by 
		Alexei V. Filippenko.

		The refraction is based on a flat atmosphere and the refractivity
		values are calculated from a modified Edlen (1953) equation.

		Parameters
		----------
		l1 	: 	float
			Shortest wavelength of interest in micron.
		l2 	: 	float
			Longest wavelength of interest in micron.
		zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
		conditions 	: string
			Refers to the standard conditions of the 
			observatory. See get_conditions() for more info.
			If set to 'None', the conditions specified 
			after will be used.
		T 	: float (optional)
			Temperature in Kelvin
		p 	: float (optional)
			Atmospheric pressure in Pa
		RH 	: float (optional)
			Relative humidity
		xc 	: float (optional)
			CO2 density in parts per million
		lat : float (optional)
			Latitude of the observer in degrees
		h 	: float (optional)
			Altitude of the observer in meters
		f 	: float
			Water vapour pressure (default=None). If None, it is
			calculated automatically.

		Returns
		-------
		R : float
			Atmospheric refraction in degrees
	'''

	if conditions is not None:
		T, p, RH, xc, lat, h, _ = get_conditions(conditions).values()

	n1 = Edlen1953(l1, T=T, p=p, RH=RH, f=f)
	n2 = Edlen1953(l2, T=T, p=p, RH=RH, f=f)
	dR  = (n1 - n2) * np.tan(np.deg2rad(zenith))
	return np.degrees(dR)


def quick_refractive_index(l, conditions='STANDARD', T=288.15, p=101325, 
							RH=0.0, xc=450, lat=0, h=0):
	'''
		Function to quickly calculate the refractive index
		of atmospheric air at reference conditions, given 
		a wavelength l.

		Parameters
		----------
		l 	: float
			Wavelength in microns
		conditions: string
			Refers to the standard conditions of the 
			observatory. See get_conditions() for more info.
			If set to 'None', the conditions specified 
			after will be used.
		T 	: float (optional)
			Temperature in Kelvin
		p 	: float (optional)
			Atmospheric pressure in Pa
		RH 	: float (optional)
			Relative humidity
		xc 	: float (optional)
			CO2 density in parts per million
		lat : float (optional)
			Latitude of the observer in degrees
		h 	: float (optional)
			Altitude of the observer in meters
	
		Returns
		-------
		n 	: 	float
			Refractive index for the given conditions.
	'''
	if conditions is not None:
		T, p, RH, xc, lat, h, _ = get_conditions(conditions).values()

	# Initializing dispersion model
	at  = Observatory()

	# Calculating indices of refraction for l
	n 	= at.n_tph(l=l, T=T, p=p, RH=RH, xc=xc)
	return n


def quick_refraction(l, zenith, conditions='STANDARD', T=288.15, p=101325, 
							RH=0.0, xc=450, lat=0, h=0):
	'''
		Function to quickly calculate the refraction
		of atmospheric air at reference conditions, 
		given a wavelength l.

		Parameters
		----------
		l  	: 	float 
			Wavelength in microns
		zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
		conditions 	: string
			Refers to the standard conditions of the 
			observatory. See get_conditions() for more info.
			If set to 'None', the conditions specified 
			after will be used.
		T  	: 	float (optional)
			Temperature in Kelvin
		p  	: 	float (optional)
			Atmospheric pressure in Pa
		RH  	: 	float (optional)
			Relative humidity
		xc  : 	float (optional)
			CO2 density in parts per million
		lat : 	float (optional)
			Latitude of the observer in degrees
		h  	: 	float (optional)
			Altitude of the observer in meters
	
		Returns
		-------
		R 	: 	float
			Refraction in degrees
	'''
	if conditions is not None:
		T, p, RH, xc, lat, h, _ = get_conditions(conditions).values()

	# Initializing dispersion model
	at  = Observatory()

	# Calculating indices of refraction for l
	n 	= at.n_tph(l=l, T=T, p=p, RH=RH, xc=xc)

	# Density of the atmosphere (following CIPM-81/91 equations)
	rho = at.rho(p=p, T=T, RH=RH, xc=xc)

	# Initializing refraction model and setting the reduced height
	ref = refraction(lat, h)
	ref.setReducedHeight(p, rho)
	return ref.cassini(n, zenith)


def quick_dispersion(l1, l2, zenith, conditions='STANDARD', T=288.15, 
						p=101325, RH=0.0, xc=450, lat=0, h=0):
	'''
		Function to quickly calculate the dispersion
		of atmospheric air at reference conditions, 
		given a wavelength l.

		Parameters
		----------
		l1 	: 	float
			Shortest wavelength of interest in micron.
		l2 	: 	float
			Longest wavelength of interest in micron.
		zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
		conditions 	: string
			Refers to the standard conditions of the 
			observatory. See get_conditions() for more info.
			If set to 'None', the conditions specified 
			after will be used.
		T 	: 	float (optional)
			Temperature in Kelvin
		p 	: 	float (optional)
			Atmospheric pressure in Pa
		RH 	: 	float (optional)
			Relative humidity
		xc 	: 	float (optional)
			CO2 density in parts per million
		lat : 	float (optional)
			Latitude of the observer in degrees
		h 	: 	float (optional)
			Altitude of the observer in meters
	
		Returns
		-------
		dR 	: float
		 	Atmospheric dispersion in degrees
	'''
	if conditions is not None:
		T, p, RH, xc, lat, h, _ = get_conditions(conditions).values()

	# Initializing dispersion model
	at  = Observatory()

	# Calculating indices of refraction for l1 and l2
	n1 	= at.n_tph(l=l1, T=T, p=p, RH=RH, xc=xc)
	n2 	= at.n_tph(l=l2, T=T, p=p, RH=RH, xc=xc)

	# Density of the atmosphere (following CIPM-81/91 equations)
	rho = at.rho(p=p, T=T, RH=RH, xc=xc)

	# Initializing refraction model and setting the reduced height
	disp = dispersion(lat, h)
	disp.setReducedHeight(p, rho)
	return disp.cassini(n1, n2, zenith)


def slalib_refraction(l, zenith, conditions='STANDARD', T=288.15, p=101325, 
				RH=0.0, xc=450, lat=0, h=0, tlr=0.0065, eps=1e-8, use_pyslalib=False):
	'''
		Calculates the atmospheric refraction assuming the SLALIB
		package. SLALIB uses the same atmospheric model as ZEMAX
		based on the works of Hohenkerk & Sinclair and Seidelmann.
		They take the B&S refraction formula as a basis.
		
		The pySLALIB FORTRAN has been ported, so pyslalib is not
		needed, but can still be used. There are minor numerical
		differences, likely due to precision or declaration issues.

		Parameters
		---------- 
		l 	:	float
			Wavelength in microns
		zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
		conditions 	: string
			Refers to the standard conditions of the 
			observatory. See get_conditions() for more info.
			If set to 'None', the conditions specified 
			after will be used.
		T 	: 	float (optional)
			Temperature in Kelvin
		p 	: 	float (optional)
			Atmospheric pressure in Pa
		RH 	: 	float (optional)
			Relative humidity
		xc 	: 	float (optional)
			CO2 density in parts per million
		lat : 	float (optional)
			Latitude of the observer in degrees (default=0)
		h 	: 	float (optional)
			Altitude of the observer in meters
		tlr :	float
			Temperature lapse rate (default=0.0065)
		eps :	float
			Precision at which the calculation is aborted (default=1e-8)
		use_pyslalib : bool
			Flag that indicates that the user wants to use the pySlalib
			package. If False, the internal python port of the original
			FORTRAN code will be used. (default=False).
			
		Returns
		-------
		R 	: 	float
			Refraction in degrees
	'''

	if conditions is not None:
		T, p, RH, xc, lat, h, _ = get_conditions(conditions).values()

	# Converting units to the units that sla_refro() requires.
	_zenith = np.deg2rad(zenith) 	# From degrees to radians
	_P  	= p / 100 				# From Pa to mbar
	_phi  	= np.deg2rad(lat) 		# From degrees to radians

	R = []
	if type(l) == float or type(l) == int:
		l = [l]
	if use_pyslalib:
		try:
		    from pyslalib import slalib
		except ImportError as e:
		    print('pySLALIB package not found! Try installing it first.')
		for wl in l:
			# If using original slalib (e.g. PySLALIB)
			Refrac = slalib.sla_refro(_zenith, h, T, _P, RH, wl, phi, tlr, eps)
			R.append(np.degrees(Refrac))
	
	# If using internal slalib port
	else:
		for wl in l:
			Refrac = sla_refro(_zenith, h, T, _P, RH, wl, phi, tlr, eps)
			R.append(np.degrees(Refrac))

	if len(l) == 1:
		return R[0]
	else:
		return np.asarray(R)

def slalib_dispersion(l1, l2, zenith, conditions='STANDARD', T=288.15, p=101325, 
				RH=0.0, xc=450, lat=0, h=0, tlr=0.0065, eps=1e-8, use_pyslalib=False):
	'''
		Calculates the atmospheric refraction assuming the SLALIB
		package. SLALIB uses the same atmospheric model as ZEMAX
		based on the works of Hohenkerk & Sinclair and Seidelmann.
		They take the B&S refraction formula as a basis.
		
		The pySLALIB FORTRAN has been ported, so pyslalib is not
		needed, but can still be used. There are minor numerical
		differences, likely due to precision or declaration issues.

		Parameters
		---------- 
		l1 	: 	float
			Shortest wavelength of interest in micron.
		l2 	: 	float
			Longest wavelength of interest in micron.
		zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
		conditions 	: string
			Refers to the standard conditions of the 
			observatory. See get_conditions() for more info.
			If set to 'None', the conditions specified 
			after will be used.
		T 	: 	float (optional)
			Temperature in Kelvin
		p 	: 	float (optional)
			Atmospheric pressure in Pa
		RH 	: 	float (optional)
			Relative humidity
		xc 	: 	float (optional)
			CO2 density in parts per million
		lat : 	float (optional)
			Latitude of the observer in degrees (default=0)
		h 	: 	float (optional)
			Altitude of the observer in meters
		tlr :	float
			Temperature lapse rate (default=0.0065)
		eps :	float
			Precision at which the calculation is aborted (default=1e-8)
		use_pyslalib : bool
			Flag that indicates that the user wants to use the pySlalib
			package. If False, the internal python port of the original
			FORTRAN code will be used. (default=False).
			
		Returns
		-------
		dR 	: 	float
			Refraction in degrees
		'''
	R1 = slalib_refraction(l1, zenith, conditions, T, p, RH, xc,\
	 lat, h, tlr, eps, use_pyslalib)

	R2 = slalib_refraction(l2, zenith, conditions, T, p, RH, xc,\
	 lat, h, tlr, eps, use_pyslalib)

	dR = R1 - R2
	return dR



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Doublet Lens Para xMatrix"""

import time
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

start_time = time.time()

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0.0
P_Obj.Thickness = 10
P_Obj.Glass = "AIR"
P_Obj.Diameter = 30.0

# ______________________________________#

L1a = Kos.surf()
L1a.Rc = 9.284706570002484E+001
L1a.Thickness = 6.0
L1a.Glass = "BK7"
L1a.Diameter = 30.0
L1a.Axicon = 0

# ______________________________________#

L1b = Kos.surf()
L1b.Rc = -3.071608670000159E+001
L1b.Thickness = 3.0
L1b.Glass = "F2"
L1b.Diameter = 30

# ______________________________________#

L1c = Kos.surf()
L1c.Rc = -7.819730726078505E+001
L1c.Thickness = 9.737604742910693E+001
L1c.Glass = "AIR"
L1c.Diameter = 30

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0.0
P_Ima.Thickness = 0.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 3.0
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, L1a, L1b, L1c, P_Ima]
config_1 = Kos.Setup()

# ______________________________________#

Doblete = Kos.system(A, config_1)
Prx = Doblete.Parax(0.4)
SistemMatrix, S_Matrix, N_Matrix, a, b, c, d, EFFL, PPA, PPP, CC, N_Prec, DD = Prx
print(EFFL)

# ______________________________________#

L1a.Rc = L1a.Rc + 1
Doblete.SetData()
Prx = Doblete.Parax(0.4)
SistemMatrix, S_Matrix, N_Matrix, a, b, c, d, EFFL, PPA, PPP, CC, N_Prec, DD = Prx
print(EFFL)

# ______________________________________#

L1a.Rc = L1a.Rc + 1
Doblete.SetData()
Prx = Doblete.Parax(0.4)
SistemMatrix, S_Matrix, N_Matrix, a, b, c, d, EFFL, PPA, PPP, CC, N_Prec, DD = Prx
print(EFFL)

# ______________________________________#


L1a.Rc = L1a.Rc + 1
Doblete.SetData()
Prx = Doblete.Parax(0.4)
SistemMatrix, S_Matrix, N_Matrix, a, b, c, d, EFFL, PPA, PPP, CC, N_Prec, DD = Prx
print(EFFL)

print("=======================================")

print(SistemMatrix)
print(S_Matrix)
print( N_Matrix)
print( a)
print( b )
print( c)
print( d)
print( EFFL)
print( PPA)
print( PPP)
print( CC)
print( N_Prec)
print( DD)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Parabole Mirror Shift"""

import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

P_Obj = Kos.surf()
P_Obj.Thickness = 5.0
P_Obj.Diameter = 10
P_Obj.Drawing = 0

P = Kos.surf()
P.Thickness = 10.0
P.Diameter = 5



file = r"prism.stl"
Solid = Kos.surf()
Solid.Diameter = 20
Solid.Solid_3d_stl = file
Solid.Glass= "BK7"
Solid.DespY = -5
Solid.AxisMove = 2
Solid.Thickness = -10



P2 = Kos.surf()
P2.Thickness = -10.0
P2.Diameter = 5
P2.Glass = "AIR"

Solid2 = Kos.surf()
Solid2.Diameter = 20
Solid2.Solid_3d_stl = file
Solid2.TiltX = 180
Solid2.Glass= "BK7"
Solid2.DespY = -5
Solid2.AxisMove = 2
Solid2.Thickness = 40

P_Ima = Kos.surf()
P_Ima.Glass = "AIR"
P_Ima.Diameter = 5
P_Ima.Drawing = 1

A = [P_Obj, P, Solid, P2, Solid2, P_Ima]
configuracion_1 = Kos.Setup()
SOLID = Kos.system(A, configuracion_1)
Rays = Kos.raykeeper(SOLID)
SOLID.energy_probability = 1




AA = [P_Obj, P, P_Ima]
SOLID2 = Kos.system(AA, configuracion_1)





W= 0.65
Surf, W, AperVal, AperType = 1, W, P.Diameter, "EPD"
P = Kos.PupilCalc(SOLID2, Surf, W, AperType, AperVal)
P.Samp, P.Ptype, P.FieldX, P.FieldType = 5, "fany", 0, "angle"
x, y, z, L, M, N = P.Pattern2Field()
Kos.NsTraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)

Kos.display2d(SOLID, Rays, 0, arrow = 1)
Kos.display3d(SOLID, Rays, 1)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Parabole Mirror Shift"""

import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos


F = 300
D = 50
h = 50
k = -1

r = np.sqrt((F**2.0) + (h**2.0))
Theta=np.rad2deg(np.arctan(h/F))
nh = r*(h/F)

W =0.65
d=1
m=1
Tm = np.arcsin((m*W/(2.0*d)))
Tm = np.rad2deg(Tm)


P_Obj = Kos.surf()
P_Obj.Thickness = r
P_Obj.Diameter = D
P_Obj.Drawing = 0

OAP1 = Kos.surf()
OAP1.Rc = -2.0 * F
OAP1.k  = k
OAP1.Thickness = -r
OAP1.Glass = "MIRROR"
OAP1.Diameter = D
OAP1.ShiftY = -h
OAP1.TiltX = -Theta
OAP1.Order = 0.0
OAP1.AxisMove = 0.0

M = Kos.surf()
M.Thickness = r
M.Glass = "MIRROR"
M.Diameter = D
M.DespY = -nh
M.TiltX = Tm
M.Order = 0.0
M.AxisMove = 0.0
M.Grating_D = d
M.Diff_Ord = m
M.Grating_Angle = 0.0

OAP2 = Kos.surf()
OAP2.Rc = -2.0 * F
OAP2.k  = k
OAP2.Thickness = - r
OAP2.Glass = "MIRROR"
OAP2.Diameter = D
OAP2.ShiftY = h
OAP2.DespY = -nh * 2.0
OAP2.TiltX = Theta
OAP2.Order = 0.0
OAP2.AxisMove = 0.0

P_Ima = Kos.surf()
P_Ima.DespY = -nh * 2.0
P_Ima.TiltX = -Theta/2.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 20

A = [P_Obj, OAP1, M, OAP2, P_Ima]
configuracion_1 = Kos.Setup()

ZernTurn = Kos.system(A, configuracion_1)
Rays = Kos.raykeeper(ZernTurn)

Surf, W, AperVal, AperType = 1, W, OAP1.Diameter, "EPD"
P = Kos.PupilCalc(ZernTurn, Surf, W, AperType, AperVal)
P.Samp, P.Ptype, P.FieldX, P.FieldType = 4, "fany", 0, "height"
x, y, z, L, M, N = P.Pattern2Field()

delta=0.001
Rang= 0.01
DW = np.arange(-Rang,Rang+delta,delta)
for dw in DW:
    Kos.TraceLoop(x, y, z, L, M, N, W + dw, Rays, clean = 0)
Kos.display3d(ZernTurn, Rays, 0)


# d=1
# Tii= Theta
# Ti = np.deg2rad(Tii)
# m=1
# Tm = np.arcsin((m*W/(2.0*d)))

# Tm = np.rad2deg(Tm)
# print(Tm)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Perfect Lens"""

import time
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

import scipy


import matplotlib.pyplot as plt
def MyPlot(RK,surf, figure= "Spot", mk=["x"], col = [[0.8,0.0,0.0]]):
    fig = plt.figure(figure)
    ax = fig.add_subplot()
    i=0
    for rk in RK:
        x,y,z,l,m,n = rk.pick(surf[i], coordinates="local")
        plt.scatter(x, y,marker = mk[i], color = col[i])
        i = i + 1
    # https://matplotlib.org/3.1.1/api/markers_api.html
    ax.set_aspect('equal', adjustable='box')
    ax.set_xlabel('x (mm)')
    ax.set_ylabel('y (mm)')
    ax.set_title('Spot diagram')
    plt.show()



def R_RMS_delta(Z1, L, M, N, X0, Y0):
    X1 = ((L / N) * Z1) + X0
    Y1 = ((M / N) * Z1) + Y0
    cenX = np.mean(X1)
    cenY = np.mean(Y1)
    x1 = (X1 - cenX)
    y1 = (Y1 - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    return R_RMS

def BestFocus(X, Y, Z, L, M, N, system, mod=1):
    delta_Z = 0
    ZZ = (L, M, N, X, Y)
    v = scipy.optimize.fsolve(R_RMS_delta, delta_Z, args=ZZ)
    if mod ==1:
        system.SDT[-2].Thickness = system.SDT[-2].Thickness + v[0]
        system.SetData()
        system.SetSolid()
    return system, v[0]


def BestRMS(system,raykeeper):
    x,y,z,l,m,n = raykeeper.pick(-1 , coordinates="local")
    system, deltaZ = BestFocus(x, y, z, l, m, n, system, mod = 0)
    rms = R_RMS_delta(deltaZ,l,m,n,x,y)
    return rms


class Function2Optimize:
    def __init__(self, system, raykeeper, effl, Surf, w, AperVal, AperType):

        self.system = system
        self.w = w
        self.effl = effl
        self.raykeeper = raykeeper

        self.P = Kos.PupilCalc(Lens, Surf, w, AperType, AperVal)

        self.P.Samp, self.P.Ptype, self.P.FieldType = 25, "fanx", "angle"
        self.P.FieldX, self.P.FieldY = 0.0, 0.0
        self.ProcessPattern2Field()
        self.deltaW= 0.125

    def ProcessPattern2Field(self):
        self.x, self.y, self.z, self.L, self.M, self.N = self.P.Pattern2Field()

    def EFFL(self, V):
        self.system.SDT[1].Rc = V[0]
        self.system.SDT[2].Rc = V[1]
        self.system.SDT[3].Rc = V[2]


        self.system.SetData()

        self.system.Parax(self.w)
        EFFL0 = self.system.EFFL

        self.system.Parax(self.w + self.deltaW)
        EFFL1 = self.system.EFFL

        self.system.Parax(self.w - self.deltaW)
        EFFL2 = self.system.EFFL


        ra = EFFL0 - EFFL1
        rb = EFFL0 - EFFL2
        r = np.sqrt((ra**2) + (rb**2))


        Kos.TraceLoop(self.x, self.y, self.z, self.L, self.M, self.N, self.w, \
                      self.raykeeper, clean = 1)
        Kos.TraceLoop(self.x, self.y, self.z, self.L, self.M, self.N,\
                      self.w + self.deltaW, self.raykeeper, clean = 0)
        Kos.TraceLoop(self.x, self.y, self.z, self.L, self.M, self.N,\
                      self.w - self.deltaW, self.raykeeper, clean = 0)

        D_EFFL = EFFL0 - self.effl
        B = BestRMS(self.system, self.raykeeper)

        self.system.RestoreData()
        return [D_EFFL, r, B]




P_Obj = Kos.surf()
P_Obj.Thickness = 100
P_Obj.Diameter = 120.0

Fa = Kos.surf()
Fa.Rc = 0.0
Fa.Thickness = 12.0
Fa.Glass = "N-BK7"
Fa.Diameter = 120.0

Fa2 = Kos.surf()
Fa2.Rc = 0.0
Fa2.Thickness = 12.0
Fa2.Glass = "N-SF5"
Fa2.Diameter = 120.0

Fb = Kos.surf()
Fb.Rc = 0.0
Fb.Thickness = 1.0
Fb.Diameter = 120
Fb.Glass = "AIR"

P_Ima = Kos.surf()
P_Ima.Thickness = 0.0
P_Ima.Diameter = 1000.0
P_Ima.Name = "Image plane"
P_Ima.Nm_Pos = (-10,10)
P_Ima.Glass = "AIR"

config_1 = Kos.Setup()

A = [P_Obj, Fa, Fa2, Fb, P_Ima]
Lens = Kos.system(A, config_1)

Rays = Kos.raykeeper(Lens)
Rays2 = Kos.raykeeper(Lens)
Rays3 = Kos.raykeeper(Lens)


W = 0.55
DW = 0.125*0.707
effl = 1000

Surf=1

Lens.Parax(W)
print("Initial effective focal length: ", Lens.EFFL)

MyFun = Function2Optimize(Lens, Rays, effl, Surf, W, Fa.Diameter, "EPD")
MyFun.deltaW=DW

R0 = 700
R1 = -700
R2 = -2000


LimInf = [-100000, -100000, -100000]
LimSup = [ 100000,  100000,  100000]
b=(LimInf, LimSup)

R = scipy.optimize.least_squares(MyFun.EFFL, [R0, R1, R2],bounds=b, verbose=2)

[R0, R1, R2] = R.x

Lens.SDT[1].Rc = R0
Lens.SDT[2].Rc = R1
Lens.SDT[3].Rc = R2

Lens.SetData()
Lens.SetSolid()


Surf, W, AperVal, AperType = 1, W, Fa.Diameter, "EPD"
P = Kos.PupilCalc(Lens, Surf, W, AperType, AperVal)
P.Samp, P.Ptype, P.FieldX, P.FieldType = 7, "fanx", 0, "angle"
x, y, z, L, M, N = P.Pattern2Field()

Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)
Kos.TraceLoop(x, y, z, L, M, N, W + DW , Rays, clean = 0)
Kos.TraceLoop(x, y, z, L, M, N, W - DW , Rays, clean = 0)


x, y, z, l, m, n = Rays.pick(-1 , coordinates="local")
system, deltaZ = BestFocus(x, y, z, l, m, n, Lens, mod = 1)

x, y, z, L, M, N = P.Pattern2Field()
Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)
Kos.TraceLoop(x, y, z, L, M, N, W + DW , Rays, clean = 0)
Kos.TraceLoop(x, y, z, L, M, N, W - DW , Rays, clean = 0)

x, y, z, l, m, n = Rays.pick(-1 , coordinates="local")
rms = R_RMS_delta(0,l,m,n,x,y)

print("R0: ", R0, " R1: ", R1, " R2: ", R2, " RMS: ", rms )

Lens.Parax(W + DW)
print("Initial effective focal length: ", Lens.EFFL)
Lens.Parax(W)
print("Initial effective focal length: ", Lens.EFFL)
Lens.Parax(W - DW)
print("Initial effective focal length: ", Lens.EFFL)



Kos.display2d(Lens, Rays, 1)





Surf, W, AperVal, AperType = 1, W, Fa.Diameter, "EPD"
P = Kos.PupilCalc(Lens, Surf, W, AperType, AperVal)
P.Samp, P.Ptype, P.FieldX, P.FieldType = 8, "hexapolar", 0, "angle"
x, y, z, L, M, N = P.Pattern2Field()

Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)
Kos.TraceLoop(x, y, z, L, M, N, W+DW, Rays, clean = 0)
Kos.TraceLoop(x, y, z, L, M, N, W-DW, Rays, clean = 0)

Lens.Parax(W)
print("Final effective focal length: ", Lens.EFFL)


x, y, z, l, m, n = Rays.pick(-1 , coordinates="local")
system, deltaZ = BestFocus(x, y, z, l, m, n, Lens, mod = 1)

x, y, z, l, m, n = Rays.pick(-1 , coordinates="local")
rms = R_RMS_delta(0,l,m,n,x,y)
print("Final RMS radius: ", rms)

x, y, z, L, M, N = P.Pattern2Field()
Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)
Kos.TraceLoop(x, y, z, L, M, N, W+DW, Rays2, clean = 1)
Kos.TraceLoop(x, y, z, L, M, N, W-DW, Rays3, clean = 1)



RAYS = [Rays, Rays2, Rays3]
MK = ["x", "*", "+"]
COL = [[0.8,0.0,0.0], [0.0,0.8,0.0], [0.0,0.0,0.8]]
SURF = [-1, -1, -1]
MyPlot(RAYS, SURF, figure= "Spot", mk = MK, col = COL)







# rays = Trace_Array(Lens, 0.5)

# x,y,z,l,m,n = rays.pick(-1, coordinates="local")
# bf = BestFocus(x, y, z, l, m, n)
# Lens.SDT[2].Thickness = Lens.SDT[2].Thickness + bf
# Lens.SetData()
# Lens.SetSolid()

# rays = Trace_Array(Lens, 0.5)
# Kos.display3d(Lens, rays, 0)
# x,y,z,l,m,n = rays.pick(-1, coordinates="local")
# MyPlot(x,y)

# Lens.Parax(0.5)
# print("Final effective focal length: ", Lens.EFFL)
# print("Final RMS radius in best focus: ", BestRMS(Lens, Wave))




import numpy as np
import KrakenOS as Kos
import scipy
import copy













def Phase(PUPIL):
    """Phase.

    Parameters
    ----------
    PUPIL :
        PUPIL
    """


#############################################################
    """ Crea una pupila con las propiedades pasadas como parámetro"""


    # alpha = abcd[0]
    # betha = abcd[1]
    # gamma = abcd[2]
    # ro = abcd[3]


    SYSTEM = PUPIL.SYSTEM
    sup = PUPIL.Surf
    W = PUPIL.W
    ApType = PUPIL.ApertureType
    ApVal = PUPIL.ApertureValue
    configuracion_1 = SYSTEM.SETUP
    W = W

    """ Se calcula la pupila """
    Pup = Kos.PupilCalc(SYSTEM, sup, W, ApType, ApVal)

    [PIX, PIY, PIZ] = PUPIL.PosPupInp
    [POX, POY, POZ] = PUPIL.PosPupOutFoc

    [PIL, PIM, PIN] = PUPIL.DirPupSal
    [POL, POM, PON] = PUPIL.DirPupSal

    Pup.Samp = PUPIL.Samp
    Pup.Ptype = PUPIL.Ptype
    Pup.FieldY = PUPIL.FieldY
    Pup.FieldX = PUPIL.FieldX
    Pup.FieldType = PUPIL.FieldType

    """se generan los rayos para trazar por la pupila"""
    x, y, z, L, M, N = Pup.Pattern2Field()





    # Pup.Samp = 1
    Pup.Ptype = 'chief'
    xc, yc, zc, Lc, Mc, Nc = Pup.Pattern2Field()

    pSource_0_c = [xc[0], yc[0], zc[0]]
    dCos_c = [Lc[0], Mc[0], Nc[0]]

    """ Traza el rayo principal"""
    SYSTEM.IgnoreVignetting()
    SYSTEM.Trace(pSource_0_c, dCos_c, W)

    """ Toma las coordenadas y los cosenos directores del rayo
    principal a la salida del sistema"""

    [Xcc, Ycc, Zcc] = SYSTEM.XYZ[(- 1)]
    [Lcc, Mcc, Ncc] = SYSTEM.LMN[(- 1)]


    resf_out =  np.sqrt(((POX - Xcc)**2.0) + ((POY- Ycc)**2.0) + (POZ**2.0))

# - - - - - - - - - - - - - - - - - - - - - - - - -
    ArrSup = PUPIL.SYSTEM.SDT

    SS = []
    for ii in range(0,len(ArrSup)):
        ArrSup[ii].EraseVTK()
        ss = copy.deepcopy(ArrSup[ii])
        SS.append(ss)

    DisPup = SS[0].Thickness

# - - - - - - - - - - - - - - - - - - - - - -
    """ No mover """


    """ No mover """
    INP_P = Kos.surf()
    INP_P.Rc = 0
    INP_P.Thickness = DisPup
    INP_P.Diameter = PUPIL.RadPupInp * 3.0
    INP_P.Glass = 'AIR'
    INP_P.DespX = xc[0]
    INP_P.DespY = yc[0]
    INP_P.TiltX = np.rad2deg(np.arcsin((- Mc[0])))
    INP_P.TiltY = np.rad2deg(np.arcsin((Lc[0] / np.cos(np.arcsin((- Mc[0]))))))
    INP_P.AxisMove = 0
    INP_P.Order = 1

# - - - - - - - - - - - - - - - - - - - - - -

    OUT_P = Kos.surf()
    OUT_P.Rc = -np.sign(POZ)*np.abs(resf_out) #+ro
    OUT_P.Thickness = 0
    OUT_P.Diameter = PUPIL.RadPupOut * 3
    OUT_P.Glass = 'AIR'
    OUT_P.DespX = POX #+ alpha
    OUT_P.DespY = POY #+ betha
    OUT_P.AxisMove = 0
    OUT_P.Order = 1

    OUT_P.TiltX = np.rad2deg(np.arcsin((- Mcc)))
    OUT_P.TiltY = np.rad2deg(np.arcsin((Lcc / np.cos(np.arcsin((- Mcc))))))

    Test = Kos.surf()
    Test.Diameter = PUPIL.RadPupOut * 3

# - - - - - - - - - - - - - - - - - - - - - - - - -

    SS.insert(1, INP_P)
    SS.append(OUT_P)


    SS[0].Thickness = 0.0
    SS[-2].Thickness = POZ #+ gamma


    SYSTEM = Kos.system(SS, configuracion_1)

# - - - - - - - - - - - - - - - - - - - - - - - - -

    """ Se configuran los contenedores de rayos"""
    RR = Kos.raykeeper(SYSTEM)

# - - - - - - - - - - - - - - - - - - - - - - - - -

    """ Configura al sistema para ignorar los viñeteos """
    SYSTEM.IgnoreVignetting()

# - - - - - - - - - - - - - - - - - - - - - - - - -

    """ Traza los rayos por systema"""

    for i in range(0, len(x)):
        """ Traza los rayos al diametro de la pupila"""
        pSource_0 = [x[i], y[i], z[i]]
        dCos = [L[i], M[i], N[i]]
        SYSTEM.Trace(pSource_0, dCos, W)
        RR.push()

# - - - - - - - - - - - - - - - - - - - - - - - - -

    All_Rays = np.asarray(RR.OP)

    axis = 1


    AR = np.sum(All_Rays,axis) - All_Rays[:,0] - (All_Rays[:,-1]*2.)

# - - - - - - - - - - - - - - - - - - - - - - - - -

    """ Traza el rayo principal"""

    SYSTEM.Trace(pSource_0_c, dCos_c, W)

    Chief = np.asarray(SYSTEM.OP)

    CH = np.sum(Chief) - Chief[0] - (Chief[-1]*2.)

# - - - - - - - - - - - - - - - - - - - - - - - - -

    Wi = (CH - AR )


# - - - - - - - - - - - - - - - - - - - - - - - - -

    XPUP = x - xc[0]
    YPUP = y - yc[0]


    x0, y0, z0, k0, l0, mi = RR.pick(1)
    xi, yi, zi, ki, li, mi = RR.pick(-1)



    # V = np.sum((Wi - (ki * alpha) - (li * betha) - gamma - ((mi - 1) * ro))**2.0)/ (len(Wi))

    # A1 = np.sum(ki ** 2.0)
    # A2 = np.sum(ki * li)
    # A3 = np.sum(ki)
    # A4 = np.sum(ki * (mi - 1.0))

    # B1 = np.sum(li * ki)
    # B2 = np.sum(li * li)
    # B3 = np.sum(li)
    # B4 = np.sum(li * (mi - 1.0))

    # C1 = np.sum(ki)
    # C2 = np.sum(li)
    # C3 = len(xi)
    # C4 = np.sum((mi - 1.0))

    # D1 = np.sum((mi - 1.0) * ki)
    # D2 = np.sum((mi - 1.0) * li)
    # D3 = np.sum((mi - 1.0))
    # D4 = np.sum((mi - 1.0) ** 2.0)



    # d1 = np.sum(Wi * ki)
    # d2 = np.sum(Wi * li)
    # d3 = np.sum(Wi)
    # d4 = np.sum(Wi * (mi - 1.0))


    # A = np.array([
    #     [A1, B1, C1, D1],
    #     [A2, B2, C2, D2],
    #     [A3, B3, C3, D3],
    #     [A4, B4, C4, D4]
    #     ])


    # d = np.array([d1, d2, d3, d4])

    # abcd = np.linalg.solve(A, d)


    # alpha = abcd[0]
    # betha = abcd[1]
    # gamma = abcd[2]
    # ro = abcd[3]


# -------------------------------------------------

    Wi = ((Wi * 1000.0) / W)

    P2V = (np.max(Wi) - np.min(Wi))


    RR.clean()
    RR.push()
    # Kos.display2d(SYSTEM,RR,1,0)



    return ((YPUP / Pup.RadPupInp), (XPUP / Pup.RadPupInp), Wi, np.abs(P2V))








# def Phase2(PUPIL):

#     """Phase2.
#     in development

#     Parameters
#     ----------
#     PUPIL :
#         PUPIL
#     """


# #############################################################
#     """ Crea una pupila con las propiedades pasadas como parámetro"""


#     # alpha = abcd[0]
#     # betha = abcd[1]
#     # gamma = abcd[2]
#     # ro = abcd[3]


#     SYSTEM = PUPIL.SYSTEM
#     sup = PUPIL.Surf
#     W = PUPIL.W
#     ApType = PUPIL.ApertureType
#     ApVal = PUPIL.ApertureValue
#     configuracion_1 = SYSTEM.SETUP
#     W = W

#     """ Se calcula la pupila """
#     Pup = Kos.PupilCalc(SYSTEM, sup, W, ApType, ApVal)

#     [PIX, PIY, PIZ] = PUPIL.PosPupInp
#     [POX, POY, POZ] = PUPIL.PosPupOutFoc

#     [PIL, PIM, PIN] = PUPIL.DirPupSal
#     [POL, POM, PON] = PUPIL.DirPupSal

#     Pup.Samp = PUPIL.Samp
#     Pup.Ptype = PUPIL.Ptype
#     Pup.FieldY = PUPIL.FieldY
#     Pup.FieldX = PUPIL.FieldX
#     Pup.FieldType = PUPIL.FieldType

#     """se generan los rayos para trazar por la pupila"""
#     x, y, z, L, M, N = Pup.Pattern2Field()


#     # Pup.Samp = 1
#     Pup.Ptype = 'chief'
#     xc, yc, zc, Lc, Mc, Nc = Pup.Pattern2Field()

#     pSource_0_c = [xc[0], yc[0], zc[0]]
#     dCos_c = [Lc[0], Mc[0], Nc[0]]

#     """ Traza el rayo principal"""
#     SYSTEM.IgnoreVignetting()
#     SYSTEM.Trace(pSource_0_c, dCos_c, W)

#     """ Toma las coordenadas y los cosenos directores del rayo
#     principal a la salida del sistema"""

#     [Xcc, Ycc, Zcc] = SYSTEM.XYZ[(- 1)]
#     [Lcc, Mcc, Ncc] = SYSTEM.LMN[(- 1)]


#     resf_out =  np.sqrt(((POX - Xcc)**2.0) + ((POY- Ycc)**2.0) + (POZ**2.0))

# # - - - - - - - - - - - - - - - - - - - - - - - - -


#     ArrSup = PUPIL.SYSTEM.SDT

#     SS = []
#     for ii in range(0,len(ArrSup)):
#         ArrSup[ii].EraseVTK()
#         ss = copy.deepcopy(ArrSup[ii])
#         SS.append(ss)

#     DisPup = SS[0].Thickness

# # - - - - - - - - - - - - - - - - - - - - - -
#     """ No mover """
#     SS[0].Thickness = 0.0
#     SS[-1].Thickness = np.sign(POZ)*np.abs(resf_out) #+ gamma

# # - - - - - - - - - - - - - - - - - - - - - -

#     """ No mover """
#     INP_P = Kos.surf()
#     INP_P.Rc = 0
#     INP_P.Thickness = DisPup
#     INP_P.Diameter = PUPIL.RadPupInp * 3.0
#     INP_P.Glass = 'AIR'
#     INP_P.DespX = xc[0]
#     INP_P.DespY = yc[0]
#     INP_P.TiltX = np.rad2deg(np.arcsin((- Mc[0])))
#     INP_P.TiltY = np.rad2deg(np.arcsin((Lc[0] / np.cos(np.arcsin((- Mc[0]))))))
#     INP_P.AxisMove = 0
#     INP_P.Order = 1

# # - - - - - - - - - - - - - - - - - - - - - -

#     OUT_P = Kos.surf()
#     OUT_P.Rc = -np.sign(POZ)*np.abs(resf_out) #+ ro
#     OUT_P.Thickness = 0
#     OUT_P.Diameter = PUPIL.RadPupOut * 3
#     OUT_P.Glass = 'AIR'
#     OUT_P.DespX = Xcc #+ alpha
#     OUT_P.DespY = Ycc #+ betha
#     OUT_P.AxisMove = 0
#     OUT_P.Order = 1


#     Test = Kos.surf()
#     Test.Diameter = PUPIL.RadPupOut * 3

# # - - - - - - - - - - - - - - - - - - - - - - - - -


#     SS.insert(1, INP_P)
#     SS.append(OUT_P)
#     SYSTEM = Kos.system(SS, configuracion_1)

# # - - - - - - - - - - - - - - - - - - - - - - - - -

#     """ Se configuran los contenedores de rayos"""
#     RR = Kos.raykeeper(SYSTEM)

# # - - - - - - - - - - - - - - - - - - - - - - - - -

#     """ Configura al sistema para ignorar los viñeteos """
#     SYSTEM.IgnoreVignetting()

# # - - - - - - - - - - - - - - - - - - - - - - - - -

#     """ Traza los rayos por systema"""

#     for i in range(0, len(x)):
#         """ Traza los rayos al diametro de la pupila"""
#         pSource_0 = [x[i], y[i], z[i]]
#         dCos = [L[i], M[i], N[i]]
#         SYSTEM.Trace(pSource_0, dCos, W)
#         RR.push()

# # - - - - - - - - - - - - - - - - - - - - - - - - -

#     All_Rays = np.asarray(RR.OP)

#     axis = 1
#     AR = (np.sum(All_Rays,axis) - All_Rays[:,0]) - All_Rays[:,-1]*2


# # - - - - - - - - - - - - - - - - - - - - - - - - -

#     """ Traza el rayo principal"""

#     SYSTEM.Trace(pSource_0_c, dCos_c, W)

#     Chief = np.asarray(SYSTEM.OP)

#     CH = (np.sum(Chief) - Chief[0]) - Chief[-1]*2

# # - - - - - - - - - - - - - - - - - - - - - - - - -

#     Wi = CH - AR

# # - - - - - - - - - - - - - - - - - - - - - - - - -

#     XPUP = x - xc[0]
#     YPUP = y - yc[0]




# # -------------------------------------------------



#     x0, y0, z0, k0, l0, mi = RR.pick(1)
#     xi, yi, zi, ki, li, mi = RR.pick(-1)
#     # V = np.sum((Wi - (ki * alpha) - (li * betha) - gamma - ((mi - 1) * ro))**2.0)/ (len(Wi))

#     # A1 = np.sum(ki ** 2.0)
#     # A2 = np.sum(ki * li)
#     # A3 = np.sum(ki)
#     # A4 = np.sum(ki * (mi - 1.0))

#     # B1 = np.sum(li * ki)
#     # B2 = np.sum(li * li)
#     # B3 = np.sum(li)
#     # B4 = np.sum(li * (mi - 1.0))

#     # C1 = np.sum(ki)
#     # C2 = np.sum(li)
#     # C3 = len(xi)
#     # C4 = np.sum((mi - 1.0))

#     # D1 = np.sum((mi - 1.0) * ki)
#     # D2 = np.sum((mi - 1.0) * li)
#     # D3 = np.sum((mi - 1.0))
#     # D4 = np.sum((mi - 1.0) ** 2.0)



#     # d1 = np.sum(Wi * ki)
#     # d2 = np.sum(Wi * li)
#     # d3 = np.sum(Wi)
#     # d4 = np.sum(Wi * (mi - 1.0))


#     # A = np.array([
#     #     [A1, B1, C1, D1],
#     #     [A2, B2, C2, D2],
#     #     [A3, B3, C3, D3],
#     #     [A4, B4, C4, D4]
#     #     ])


#     # d = np.array([d1, d2, d3, d4])

#     # abcd = np.linalg.solve(A, d)





#     # alpha = abcd[0]
#     # betha = abcd[1]
#     # gamma = abcd[2]
#     # ro = abcd[3]



# # -------------------------------------------------

#     Wi = ((Wi * 1000.0) / W)




#     P2V = (np.max(Wi) - np.min(Wi))

#     RR.clean()
#     RR.push()
#     # Kos.display2d(SYSTEM,RR,1,0)


#     # Wi = Wi - np.median(Wi)

#     return ((YPUP / Pup.RadPupInp), (XPUP / Pup.RadPupInp), Wi, np.abs(P2V))





# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Tel 2M Wavefront Fitting"""

import os
import sys
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources

""" Looking for if KrakenOS is installed, if not, it assumes that
an folder downloaded from github is run"""

required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("Not installed")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

currentDirectory = os.getcwd()
sys.path.insert(1, currentDirectory + '/library')

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0
P_Obj.Thickness = 1000*0 + 3452.2
P_Obj.Glass = "AIR"
P_Obj.Diameter = 1059. * 2.0
P_Obj.Drawing=0

# ______________________________________#

Thickness = 3452.2
M1 = Kos.surf()
M1.Rc = -9638.0
M1.Thickness = -Thickness
M1.k = -1.07731
M1.Glass = "MIRROR"
M1.Diameter = 1.059E+003 * 2.0
M1.InDiameter = 250 * 2.0
M1.TiltY = 0.0
M1.TiltX = 0.0
M1.AxisMove = 0
# ______________________________________#

M2 = Kos.surf()
M2.Rc = -3930.0
M2.Thickness = Thickness + 1037.525880
M2.k = -4.3281
M2.Glass = "MIRROR"
M2.Diameter = 336.5 * 2.0
M2.TiltY = 0.0
M2.TiltX = 0.0
M2.DespY = 0.0
""" Se inclina el secundario """
M2.DespX = 1.0
M2.AxisMove = 0

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Diameter = 300.0
P_Ima.Glass = "AIR"
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, M1, M2, P_Ima]
configuracion_1 = Kos.Setup()
Telescopio = Kos.system(A, configuracion_1)

# ______________________________________#
"""Vamos a definir los parámetros de la pupila del sistema,
definiremos esta pupila en la superficie 1, esta corresponde al
espejo primario"""
Surf = 1

"""Definimos la longitud de onda en micras"""
W = 0.50169

"""Dindicamos que la apertura del sistema definirá la pupila"""
AperType = "EPD"

""" Definimos el Diametro de la apertura del sistema"""
AperVal = 2000.
Pupil = Kos.PupilCalc(Telescopio, Surf, W, AperType, AperVal)




print("Radio pupila de entrada: ")
print(Pupil.RadPupInp)
print("Posición pupila de entrada: ")
print(Pupil.PosPupInp)
print("Rádio pupila de salida: ")
print(Pupil.RadPupOut)
print("Posicion pupila de salida: ")
print(Pupil.PosPupOut)
print("Posicion pupila de salida respecto al plano focal: ")
print(Pupil.PosPupOutFoc)
print("Orientación pupila de salida")
print(Pupil.DirPupSal)

print("Airy disk radius focal distance (micrometers)")
print(Pupil.FocusAiryRadius)


print("Distancia focal")
print(Pupil.EFFL)







""" Para los calculos internos de la fase del frente de onda indicamos que
la pupila tendrá un arreglo exapolar con 10 anillos"""

Pupil.Samp = 11
Pupil.Ptype = "hexapolar"
"""Indicamos que los campos son tel tipo angulo, como en el caso de los telescopios
con luz desde el infinito, para diseños con objeto certano este parametro es la altura
 del objeto"""

Pupil.FieldType = "angle"
""" Definimos que el campo es 0 en x y cero en y, es decir, está en el eje óptico"""


Pupil.FieldX = 0.1
Pupil.FieldY = 0.0


""" Ahora calculamos la fase del frente de onda en la pupila, las coordenadas X, Y
son las coordendas en la pupila, el valor de Z es la fase en cada punto X, Y y P2V
es el valor pico a valle."""

X, Y, Z, P2V = Kos.Phase(Pupil)
print("Peak to valley: ", P2V)


"""Indicamos el grado de expanción para los polinomios de Zernike"""
NC = 35

"""Generamos un arreglo numpy conlas mismas dimensiones de la expanción definida"""
A = np.ones(NC)

"""Calculamos los polinomios de Zernike con la fase calculada y el numero de
 elementos deseados en la expanción, Zcoef son los coeficientes en longitudes de
 onda, Mat es la expreción matematica de Zeidel para dicho coeficiente,
 esto con fines ilustrativos, w_rms es el error del ajuste"""

Zcoef, Mat, RMS2Chief, RMS2Centroid, FITTINGERROR = Kos.Zernike_Fitting(X, Y, Z, A)

""""Se despliegan los resultados"""
for i in range(0, NC):
    print("z", i + 1, "  ", "{0:.8f}".format(float(Zcoef[i])), ":", Mat[i])


# ______________________________________#
print("(RMS) Fitting error: ", FITTINGERROR)
print(RMS2Chief, "RMS(to chief) From fitted coefficents")
print(RMS2Centroid, "RMS(to centroid) From fitted coefficents")



COEF = Zcoef
Focal = Pupil.EFFL
Diameter = 2.0 * Pupil.RadPupInp
Wave = W
I= Kos.psf(COEF, Focal, Diameter, Wave,pixels=265, plot=1)



"""Se genera un contenedor de rayos"""

RR = Kos.raykeeper(Telescopio)

""" Se generan rayos que pasan por la pupila con la configuración realizada antes"""
x, y, z, L, M, N = Pupil.Pattern2Field()

# ______________________________________#

""" Se trazan esos rayos y se almacenan"""

for i in range(0, len(x)):
    pSource_0 = [x[i], y[i], z[i]]
    dCos = [L[i], M[i], N[i]]
    Telescopio.Trace(pSource_0, dCos, W)
    RR.push()

# ______________________________________#

""" Se grafica el telescopio con los rayos almacenados"""

Kos.display3d(Telescopio, RR, 1 )
X, Y, Z, L, M, N = RR.pick(-1)

# ______________________________________#

""" Se grafica el diagrama de manchas """
plt.figure(2)
plt.plot(X, Y, 'x')
plt.xlabel('X')
plt.ylabel('Y')
plt.title('spot Diagram')
plt.axis('square')

K = .100
Lx = np.mean(X)
Ly = np.mean(Y)
left = (-K) + Lx
right = (K) + Lx
up = (K) + Ly
down = (-K) + Ly

plt.xlim([left, right])
plt.ylim([up, down])


plt.show()





X = X - np.mean(X)
Y = Y - np.mean(Y)

R=np.sqrt(X**2 + Y**2)
RMS = np.sqrt(np.mean(R**2))
print("RMS Radius(mm): ", RMS)

""" Se prepara una imagen con los coeficientes de 400x400"""
# Zcoef[0]=0.
# Zcoef[1]=0.
# Zcoef[2]=0.
ima = Kos.WavefrontData2Image(Zcoef, 400)

print("Peak 2 valley. ", np.max(ima)-np.min(ima))



""" Se grafica el interferograma """
Type = "interferogram"

# ima = np.flipud(ima)
Kos.ZernikeDataImage2Plot(ima, Type)


# """ Se calculan las sumas de Seidel """
# AB = Kos.Seidel(Pupil)

# print("--------------------------------------")
# print(AB.SCW_AN)
# print(AB.SCW_NM)
# print(AB.SCW_TOTAL)






# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Tel 2M Wavefront Fitting"""

import os
import sys
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
""" Looking for if KrakenOS is installed, if not, it assumes that
an library downloaded from github is running"""

required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("Not installed")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

import scipy



###########-----------------------




import matplotlib.pyplot as plt
def MyPlot(RK,surf, figure= "Spot", mk=["x"], col = [[0.8,0.0,0.0]]):
    fig = plt.figure(figure)
    ax = fig.add_subplot()
    i=0
    for rk in RK:
        x,y,z,l,m,n = rk.pick(surf[i], coordinates="local")
        plt.scatter(x, y,marker = mk[i], color = col[i])
        i = i + 1
    # https://matplotlib.org/3.1.1/api/markers_api.html
    ax.set_aspect('equal', adjustable='box')
    ax.set_xlabel('x (mm)')
    ax.set_ylabel('y (mm)')
    ax.set_title('Spot diagram')
    plt.show()



def R_RMS_delta(Z1, L, M, N, X0, Y0):
    X1 = ((L / N) * Z1) + X0
    Y1 = ((M / N) * Z1) + Y0
    cenX = np.mean(X1)
    cenY = np.mean(Y1)
    x1 = (X1 - cenX)
    y1 = (Y1 - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    return R_RMS


def BestRMS_2(system,raykeeper):
    x,y,z,l,m,n = raykeeper.pick(-1 , coordinates="local")
    rms = R_RMS_delta(0,l,m,n,x,y)
    return rms


class Function2Optimize:
    def __init__(self, system, raykeeper, effl, Surf, w, AperVal, AperType):

        self.system = system
        self.w = w
        self.effl = effl
        self.raykeeper = raykeeper

        self.P = Kos.PupilCalc(system, Surf, w, AperType, AperVal)

        self.P.Samp, self.P.Ptype, self.P.FieldType = 25, "fanx", "angle"
        self.P.FieldX, self.P.FieldY = 0.0, 0.0
        self.ProcessPattern2Field()

    def ProcessPattern2Field(self):
        self.x, self.y, self.z, self.L, self.M, self.N = self.P.Pattern2Field()

    def EFFL(self, V):
        self.system.SDT[1].Rc = V[0]
        self.system.SDT[1].Thickness = V[1]
        self.system.SDT[2].Thickness = -V[1]
        self.system.SDT[2].Rc = V[0]
        self.system.SetData()

        self.system.Parax(self.w)
        D_EFFL = self.system.EFFL - self.effl

        Kos.TraceLoop(self.x, self.y, self.z, self.L, self.M, self.N, self.w, \
                      self.raykeeper, clean = 1)

        B = BestRMS_2(self.system, self.raykeeper)
        self.system.RestoreData()
        return [D_EFFL, B]


# ______________________________________#

currentDirectory = os.getcwd()
sys.path.insert(1, currentDirectory + '/library')

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Thickness = 2000.0
P_Obj.Glass = "AIR"
P_Obj.Diameter = 1300.0
P_Obj.Drawing = 0

Thickness = 1778.0
M1 = Kos.surf()
M1.Rc = -6081.3
M1.Thickness = -Thickness
M1.k = -1.0
M1.Glass = "MIRROR"
M1.Diameter = 1300
M1.InDiameter = 650.0 * 0.5

M2 = Kos.surf()
M2.Rc = -6081.3
M2.Thickness = Thickness
M2.k = 0.0
M2.Glass = "MIRROR"
M2.Diameter = 650.0

M1Vertex=Kos.surf(Thickness = 50.0, Glass = "AIR",Diameter = 600.0)
M1Vertex.Drawing=0

Corr_1=Kos.surf(Thickness = 10, Glass = "LITHOSIL-Q", Diameter=230)

ED1=np.zeros(20)
ED1[0] = 0.0
ED1[1] = 0.0
Corr_1.AspherData = ED1

Corr_2=Kos.surf(Thickness=390, Glass="AIR", Diameter=230)
Corr_2.AspherData = -ED1

P_Ima = Kos.surf()
P_Ima.Diameter = 300.0

A = [P_Obj, M1, M2, M1Vertex, Corr_1, Corr_2, P_Ima]
configuracion_1 = Kos.Setup()
Telescope = Kos.system(A, configuracion_1)
# ______________________________________#


Rays = Kos.raykeeper(Telescope)


W = 0.6
effl = 5200
Surf=1

MyFun = Function2Optimize(Telescope, Rays, effl, Surf, W, M1.Diameter, "EPD")


R0 = -6081.3
D  = -1778.0


LimInf = [-6081.3 - 100, -1778.0 - 100]
LimSup = [-6081.3 + 100, -1778.0 + 100]
b=(LimInf, LimSup)

R = scipy.optimize.least_squares(MyFun.EFFL, [R0, D],bounds=b, verbose=2)

[R0, D] = R.x

Telescope.SDT[1].Rc = R0
Telescope.SDT[1].Thickness = D
Telescope.SDT[2].Thickness = -D
Telescope.SDT[2].Rc = R0
Telescope.SetData()
Telescope.SetSolid()


Surf, W, AperVal, AperType = 1, W, M1.Diameter, "EPD"
P = Kos.PupilCalc(Telescope, Surf, W, AperType, AperVal)
P.Samp, P.Ptype, P.FieldX, P.FieldType = 7, "fany", 0, "angle"
x, y, z, L, M, N = P.Pattern2Field()

Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)
x, y, z, l, m, n = Rays.pick(-1 , coordinates="local")

rms = R_RMS_delta(0,l,m,n,x,y)
print("R0: ", R0, " R1: ", -R0, " D: ", D, " RMS: ", rms )

Kos.display3d(Telescope, Rays, 2)



Telescope.Parax(W)
print(Telescope.EFFL)


P.Samp, P.Ptype, P.FieldX, P.FieldType = 7, "hexapolar", 0, "angle"
x, y, z, L, M, N = P.Pattern2Field()
Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)
RAYS = [Rays]
MK = ["x"]
COL = [[0.8,0.0,0.0]]
SURF = [-1]
MyPlot(RAYS, SURF, figure= "Spot2", mk = MK, col = COL)





import numpy as np
import pyvista as pv

class raykeeper():
    """raykeeper.
    """


    def __init__(self, System):
        """__init__.

        Parameters
        ----------
        System :
            System
        """
        self.SYSTEM = System
        self.clean()

    def valid(self):
        """valid.
        """
        z = np.argwhere((self.vld == 1))
        return z

    def push(self):
        """push.
        """
        self.nelements = self.SYSTEM.n
        if (self.SYSTEM.val == 0):
            self.invalid_vld = np.append(self.vld, 0)
            self.invalid_SURFACE.append(np.asarray(self.SYSTEM.SURFACE))
            self.invalid_NAME.append(np.asarray(self.SYSTEM.NAME))
            self.invalid_GLASS.append(np.asarray(self.SYSTEM.GLASS))
            self.invalid_S_XYZ.append(np.asarray(self.SYSTEM.S_XYZ))
            self.invalid_T_XYZ.append(np.asarray(self.SYSTEM.T_XYZ))
            self.invalid_XYZ.append(np.asarray(self.SYSTEM.XYZ))
            self.invalid_OST_XYZ.append(np.asarray(self.SYSTEM.OST_XYZ))
            self.invalid_OST_LMN.append(np.asarray(self.SYSTEM.OST_LMN))
            self.invalid_S_LMN.append(np.asarray(self.SYSTEM.S_LMN))
            self.invalid_LMN.append(np.asarray(self.SYSTEM.LMN))
            self.invalid_R_LMN.append(np.asarray(self.SYSTEM.R_LMN))
            self.invalid_N0.append(np.asarray(self.SYSTEM.N0))
            self.invalid_N1.append(np.asarray(self.SYSTEM.N1))
            self.invalid_WAV.append(np.asarray(self.SYSTEM.WAV))
            self.invalid_G_LMN.append(np.asarray(self.SYSTEM.G_LMN))
            self.invalid_ORDER.append(np.asarray(self.SYSTEM.ORDER))
            self.invalid_GRATING.append(np.asarray(self.SYSTEM.GRATING))
            self.invalid_DISTANCE.append(np.asarray(self.SYSTEM.DISTANCE))
            self.invalid_OP.append(np.asarray(self.SYSTEM.OP))
            self.invalid_TOP_S.append(np.asarray(self.SYSTEM.TOP_S))
            self.invalid_TOP.append(np.asarray(self.SYSTEM.TOP))
            self.invalid_ALPHA.append(np.asarray(self.SYSTEM.ALPHA))
            self.invalid_BULK_TRANS.append(np.asarray(self.SYSTEM.BULK_TRANS))
            self.invalid_RP.append(np.asarray(self.SYSTEM.RP))
            self.invalid_RS.append(np.asarray(self.SYSTEM.RS))
            self.invalid_TP.append(np.asarray(self.SYSTEM.TP))
            self.invalid_TS.append(np.asarray(self.SYSTEM.TS))
            self.invalid_TTBE.append(np.asarray(self.SYSTEM.TTBE))
            self.invalid_TT.append(np.asarray(self.SYSTEM.TT))
        else:
            self.vld = np.append(self.vld, 1)
            self.valid_vld = np.append(self.vld, 0)
            self.valid_SURFACE.append(np.asarray(self.SYSTEM.SURFACE))
            self.valid_NAME.append(np.asarray(self.SYSTEM.NAME))
            self.valid_GLASS.append(np.asarray(self.SYSTEM.GLASS))
            self.valid_S_XYZ.append(np.asarray(self.SYSTEM.S_XYZ))
            self.valid_T_XYZ.append(np.asarray(self.SYSTEM.T_XYZ))
            self.valid_XYZ.append(np.asarray(self.SYSTEM.XYZ))
            self.valid_OST_XYZ.append(np.asarray(self.SYSTEM.OST_XYZ))
            self.valid_OST_LMN.append(np.asarray(self.SYSTEM.OST_LMN))
            self.valid_S_LMN.append(np.asarray(self.SYSTEM.S_LMN))
            self.valid_LMN.append(np.asarray(self.SYSTEM.LMN))
            self.valid_R_LMN.append(np.asarray(self.SYSTEM.R_LMN))
            self.valid_N0.append(np.asarray(self.SYSTEM.N0))
            self.valid_N1.append(np.asarray(self.SYSTEM.N1))
            self.valid_WAV.append(np.asarray(self.SYSTEM.WAV))
            self.valid_G_LMN.append(np.asarray(self.SYSTEM.G_LMN))
            self.valid_ORDER.append(np.asarray(self.SYSTEM.ORDER))
            self.valid_GRATING.append(np.asarray(self.SYSTEM.GRATING))
            self.valid_DISTANCE.append(np.asarray(self.SYSTEM.DISTANCE))
            self.valid_OP.append(np.asarray(self.SYSTEM.OP))
            self.valid_TOP_S.append(np.asarray(self.SYSTEM.TOP_S))
            self.valid_TOP.append(np.asarray(self.SYSTEM.TOP))
            self.valid_ALPHA.append(np.asarray(self.SYSTEM.ALPHA))
            self.valid_BULK_TRANS.append(np.asarray(self.SYSTEM.BULK_TRANS))
            self.valid_RP.append(np.asarray(self.SYSTEM.RP))
            self.valid_RS.append(np.asarray(self.SYSTEM.RS))
            self.valid_TP.append(np.asarray(self.SYSTEM.TP))
            self.valid_TS.append(np.asarray(self.SYSTEM.TS))
            self.valid_TTBE.append(np.asarray(self.SYSTEM.TTBE))
            self.valid_TT.append(np.asarray(self.SYSTEM.TT))
        self.nrays = (self.nrays + 1)


        self.RayWave.append(self.SYSTEM.Wave)
        self.CC.append(self.SYSTEM.ray_SurfHits)



        self.SURFACE.append(np.asarray(self.SYSTEM.SURFACE))
        self.NAME.append(np.asarray(self.SYSTEM.NAME))
        self.GLASS.append(np.asarray(self.SYSTEM.GLASS))
        self.S_XYZ.append(np.asarray(self.SYSTEM.S_XYZ))
        self.T_XYZ.append(np.asarray(self.SYSTEM.T_XYZ))
        self.XYZ.append(np.asarray(self.SYSTEM.XYZ))
        self.OST_XYZ.append(np.asarray(self.SYSTEM.OST_XYZ))
        self.OST_LMN.append(np.asarray(self.SYSTEM.OST_LMN))
        self.S_LMN.append(np.asarray(self.SYSTEM.S_LMN))
        self.LMN.append(np.asarray(self.SYSTEM.LMN))
        self.R_LMN.append(np.asarray(self.SYSTEM.R_LMN))
        self.N0.append(np.asarray(self.SYSTEM.N0))
        self.N1.append(np.asarray(self.SYSTEM.N1))
        self.WAV.append(np.asarray(self.SYSTEM.WAV))
        self.G_LMN.append(np.asarray(self.SYSTEM.G_LMN))
        self.ORDER.append(np.asarray(self.SYSTEM.ORDER))
        self.GRATING.append(np.asarray(self.SYSTEM.GRATING))
        self.DISTANCE.append(np.asarray(self.SYSTEM.DISTANCE))
        self.OP.append(np.asarray(self.SYSTEM.OP))
        self.TOP_S.append(np.asarray(self.SYSTEM.TOP_S))
        self.TOP.append(np.asarray(self.SYSTEM.TOP))
        self.ALPHA.append(np.asarray(self.SYSTEM.ALPHA))
        self.BULK_TRANS.append(np.asarray(self.SYSTEM.BULK_TRANS))
        self.RP.append(np.asarray(self.SYSTEM.RP))
        self.RS.append(np.asarray(self.SYSTEM.RS))
        self.TP.append(np.asarray(self.SYSTEM.TP))
        self.TS.append(np.asarray(self.SYSTEM.TS))
        self.TTBE.append(np.asarray(self.SYSTEM.TTBE))
        self.TT.append(np.asarray(self.SYSTEM.TT))

    def clean(self):
        """clean.
        """
        self.vld = np.asarray([])
        self.nrays = 0
        self.RayWave = []
        self.CC =[]
        self.SURFACE = []
        self.NAME = []
        self.GLASS = []
        self.S_XYZ = []
        self.T_XYZ = []
        self.XYZ = []
        self.OST_XYZ = []
        self.OST_LMN = []
        self.S_LMN = []
        self.LMN = []
        self.R_LMN = []
        self.N0 = []
        self.N1 = []
        self.WAV = []
        self.G_LMN = []
        self.ORDER = []
        self.GRATING = []
        self.DISTANCE = []
        self.OP = []
        self.TOP_S = []
        self.TOP = []
        self.ALPHA = []
        self.BULK_TRANS = []
        self.RP = []
        self.RS = []
        self.TP = []
        self.TS = []
        self.TTBE = []
        self.TT = []
        self.valid_RayWave = []
        self.valid_CCC = pv.MultiBlock()
        self.valid_SURFACE = []
        self.valid_NAME = []
        self.valid_GLASS = []
        self.valid_S_XYZ = []
        self.valid_T_XYZ = []
        self.valid_XYZ = []
        self.valid_OST_XYZ = []
        self.valid_OST_LMN = []
        self.valid_S_LMN = []
        self.valid_LMN = []
        self.valid_R_LMN = []
        self.valid_N0 = []
        self.valid_N1 = []
        self.valid_WAV = []
        self.valid_G_LMN = []
        self.valid_ORDER = []
        self.valid_GRATING = []
        self.valid_DISTANCE = []
        self.valid_OP = []
        self.valid_TOP_S = []
        self.valid_TOP = []
        self.valid_ALPHA = []
        self.valid_BULK_TRANS = []
        self.valid_RP = []
        self.valid_RS = []
        self.valid_TP = []
        self.valid_TS = []
        self.valid_TTBE = []
        self.valid_TT = []
        self.invalid_RayWave = []
        self.invalid_CCC = pv.MultiBlock()
        self.invalid_SURFACE = []
        self.invalid_NAME = []
        self.invalid_GLASS = []
        self.invalid_S_XYZ = []
        self.invalid_T_XYZ = []
        self.invalid_XYZ = []
        self.invalid_OST_XYZ = []
        self.invalid_OST_LMN = []
        self.invalid_S_LMN = []
        self.invalid_LMN = []
        self.invalid_R_LMN = []
        self.invalid_N0 = []
        self.invalid_N1 = []
        self.invalid_WAV = []
        self.invalid_G_LMN = []
        self.invalid_ORDER = []
        self.invalid_GRATING = []
        self.invalid_DISTANCE = []
        self.invalid_OP = []
        self.invalid_TOP_S = []
        self.invalid_TOP = []
        self.invalid_ALPHA = []
        self.invalid_BULK_TRANS = []
        self.invalid_RP = []
        self.invalid_RS = []
        self.invalid_TP = []
        self.invalid_TS = []
        self.invalid_TTBE = []
        self.invalid_TT = []

    def pick(self, N_ELEMENT=(- 1), coordinates = "global"):
        """pick.

        Parameters
        ----------
        N_ELEMENT :
            N_ELEMENT

            coordinates = "global" or "local"
        """

        gls = self.SYSTEM.SDT[N_ELEMENT].Glass
        if gls == "NULL":
            print("NULL surface has been chosen, the return values correspond to those of the previous surface")

        self.numsup = (self.nelements - 1)

        if coordinates == "global":
            self.xyz = self.valid_XYZ
            self.lmn = self.valid_LMN
        else:
            self.xyz = self.valid_OST_XYZ
            self.lmn = self.valid_OST_LMN

        self.s = self.valid_SURFACE
        if ((N_ELEMENT < 0) or (N_ELEMENT > self.numsup)):
            N_ELEMENT = self.numsup
        else:
            N_ELEMENT = N_ELEMENT
        AA = []
        BB = []
        for k in self.s:
            aa = np.argwhere((k == N_ELEMENT))
            aa = np.squeeze(aa)
            AA.append(aa)
            BB.append(np.size(aa))
        AA = np.asarray(AA)
        BB = np.asarray(BB)
        if (N_ELEMENT != 0):
            BB = np.argwhere((BB == 1))
        else:
            BB = np.argwhere((BB == 0))
        X = []
        Y = []
        Z = []
        L = []
        M = []
        N = []
        for c in BB:
            for d in c:
                ray0 = self.xyz[d]

                [x1, y1, z1] = ray0[N_ELEMENT]
                X.append(x1)
                Y.append(y1)
                Z.append(z1)
                ray1 = self.lmn[d]
                if (N_ELEMENT != 0):
                    el = (N_ELEMENT - 1)
                else:
                    el = 0
                [l1, m1, n1] = ray1[el]
                L.append(l1)
                M.append(m1)
                N.append(n1)
        return (np.asarray(X), np.asarray(Y), np.asarray(Z), np.asarray(L), np.asarray(M), np.asarray(N))






# import numpy as np
# import pyvista as pv



# class raykeeper():
#     """raykeeper.
#     """


#     def __init__(self, System):
#         """__init__.

#         Parameters
#         ----------
#         System :
#             System
#         """
#         self.RaysMem=300
#         self.SYSTEM = System
#         self.clean()


#     def valid(self):
#         """valid.
#         """
#         z = np.argwhere((self.vld == 1))
#         return z

#     def push(self):
#         """push.
#         """
#         self.nelements = self.SYSTEM.n
#         if (self.SYSTEM.val == 0):
#             self.invalid_vld = np.append(self.vld, 0)
#             self.invalid_SURFACE.append( (self.SYSTEM.SURFACE))
#             self.invalid_NAME.append( (self.SYSTEM.NAME))
#             self.invalid_GLASS.append( (self.SYSTEM.GLASS))
#             self.invalid_S_XYZ.append( (self.SYSTEM.S_XYZ))
#             self.invalid_T_XYZ.append( (self.SYSTEM.T_XYZ))
#             self.invalid_XYZ.append( (self.SYSTEM.XYZ))
#             self.invalid_OST_XYZ.append( (self.SYSTEM.OST_XYZ))
#             self.invalid_S_LMN.append( (self.SYSTEM.S_LMN))
#             self.invalid_LMN.append( (self.SYSTEM.LMN))
#             self.invalid_R_LMN.append( (self.SYSTEM.R_LMN))
#             self.invalid_N0.append( (self.SYSTEM.N0))
#             self.invalid_N1.append( (self.SYSTEM.N1))
#             self.invalid_WAV.append( (self.SYSTEM.WAV))
#             self.invalid_G_LMN.append( (self.SYSTEM.G_LMN))
#             self.invalid_ORDER.append( (self.SYSTEM.ORDER))
#             self.invalid_GRATING.append( (self.SYSTEM.GRATING))
#             self.invalid_DISTANCE.append( (self.SYSTEM.DISTANCE))
#             self.invalid_OP.append( (self.SYSTEM.OP))
#             self.invalid_TOP_S.append( (self.SYSTEM.TOP_S))
#             self.invalid_TOP.append( (self.SYSTEM.TOP))
#             self.invalid_ALPHA.append( (self.SYSTEM.ALPHA))
#             self.invalid_BULK_TRANS.append( (self.SYSTEM.BULK_TRANS))
#             self.invalid_RP.append( (self.SYSTEM.RP))
#             self.invalid_RS.append( (self.SYSTEM.RS))
#             self.invalid_TP.append( (self.SYSTEM.TP))
#             self.invalid_TS.append( (self.SYSTEM.TS))
#             self.invalid_TTBE.append( (self.SYSTEM.TTBE))
#             self.invalid_TT.append( (self.SYSTEM.TT))
#         else:
#             self.vld = np.append(self.vld, 1)
#             self.valid_vld = np.append(self.vld, 0)
#             self.valid_SURFACE.append( (self.SYSTEM.SURFACE))
#             self.valid_NAME.append( (self.SYSTEM.NAME))
#             self.valid_GLASS.append( (self.SYSTEM.GLASS))
#             self.valid_S_XYZ.append( (self.SYSTEM.S_XYZ))
#             self.valid_T_XYZ.append( (self.SYSTEM.T_XYZ))
#             self.valid_XYZ.append( (self.SYSTEM.XYZ))
#             self.valid_OST_XYZ.append( (self.SYSTEM.OST_XYZ))
#             self.valid_S_LMN.append( (self.SYSTEM.S_LMN))
#             self.valid_LMN.append( (self.SYSTEM.LMN))
#             self.valid_R_LMN.append( (self.SYSTEM.R_LMN))
#             self.valid_N0.append( (self.SYSTEM.N0))
#             self.valid_N1.append( (self.SYSTEM.N1))
#             self.valid_WAV.append( (self.SYSTEM.WAV))
#             self.valid_G_LMN.append( (self.SYSTEM.G_LMN))
#             self.valid_ORDER.append( (self.SYSTEM.ORDER))
#             self.valid_GRATING.append( (self.SYSTEM.GRATING))
#             self.valid_DISTANCE.append( (self.SYSTEM.DISTANCE))
#             self.valid_OP.append( (self.SYSTEM.OP))
#             self.valid_TOP_S.append( (self.SYSTEM.TOP_S))
#             self.valid_TOP.append( (self.SYSTEM.TOP))
#             self.valid_ALPHA.append( (self.SYSTEM.ALPHA))
#             self.valid_BULK_TRANS.append( (self.SYSTEM.BULK_TRANS))
#             self.valid_RP.append( (self.SYSTEM.RP))
#             self.valid_RS.append( (self.SYSTEM.RS))
#             self.valid_TP.append( (self.SYSTEM.TP))
#             self.valid_TS.append( (self.SYSTEM.TS))
#             self.valid_TTBE.append( (self.SYSTEM.TTBE))
#             self.valid_TT.append( (self.SYSTEM.TT))

#         self.nrays = (self.nrays + 1)



#         self.RayWave.append(self.SYSTEM.Wave)
#         self.CC.append(self.SYSTEM.ray_SurfHits)



















#         self.SURFACE.append( (self.SYSTEM.SURFACE))
#         self.NAME.append( (self.SYSTEM.NAME))
#         self.GLASS.append( (self.SYSTEM.GLASS))
#         self.S_XYZ.append( (self.SYSTEM.S_XYZ))
#         self.T_XYZ.append( (self.SYSTEM.T_XYZ))
#         self.XYZ.append( (self.SYSTEM.XYZ))
#         self.OST_XYZ.append( (self.SYSTEM.OST_XYZ))
#         self.S_LMN.append( (self.SYSTEM.S_LMN))
#         self.LMN.append( (self.SYSTEM.LMN))
#         self.R_LMN.append( (self.SYSTEM.R_LMN))
#         self.N0.append( (self.SYSTEM.N0))
#         self.N1.append( (self.SYSTEM.N1))
#         self.WAV.append( (self.SYSTEM.WAV))
#         self.G_LMN.append( (self.SYSTEM.G_LMN))
#         self.ORDER.append( (self.SYSTEM.ORDER))
#         self.GRATING.append( (self.SYSTEM.GRATING))
#         self.DISTANCE.append( (self.SYSTEM.DISTANCE))
#         self.OP.append( (self.SYSTEM.OP))
#         self.TOP_S.append( (self.SYSTEM.TOP_S))
#         self.TOP.append( (self.SYSTEM.TOP))
#         self.ALPHA.append( (self.SYSTEM.ALPHA))
#         self.BULK_TRANS.append( (self.SYSTEM.BULK_TRANS))
#         self.RP.append( (self.SYSTEM.RP))
#         self.RS.append( (self.SYSTEM.RS))
#         self.TP.append( (self.SYSTEM.TP))
#         self.TS.append( (self.SYSTEM.TS))
#         self.TTBE.append( (self.SYSTEM.TTBE))
#         self.TT.append( (self.SYSTEM.TT))






#     def clean(self):
#         """clean.
#         """
#         self.vld = np.asarray([])
#         self.nrays = 0
#         self.RayWave = []
#         self.CCC = pv.MultiBlock()
#         self.CC = []
#         self.SURFACE = []
#         self.NAME = []
#         self.GLASS = []
#         self.S_XYZ = []
#         self.T_XYZ = []
#         self.XYZ = []
#         self.OST_XYZ = []
#         self.S_LMN = []
#         self.LMN = []
#         self.R_LMN = []
#         self.N0 = []
#         self.N1 = []
#         self.WAV = []
#         self.G_LMN = []
#         self.ORDER = []
#         self.GRATING = []
#         self.DISTANCE = []
#         self.OP = []
#         self.TOP_S = []
#         self.TOP = []
#         self.ALPHA = []
#         self.BULK_TRANS = []
#         self.RP = []
#         self.RS = []
#         self.TP = []
#         self.TS = []
#         self.TTBE = []
#         self.TT = []
#         self.valid_RayWave = []
#         self.valid_CCC = pv.MultiBlock()
#         self.valid_SURFACE = []
#         self.valid_NAME = []
#         self.valid_GLASS = []
#         self.valid_S_XYZ = []
#         self.valid_T_XYZ = []
#         self.valid_XYZ = []
#         self.valid_OST_XYZ = []
#         self.valid_S_LMN = []
#         self.valid_LMN = []
#         self.valid_R_LMN = []
#         self.valid_N0 = []
#         self.valid_N1 = []
#         self.valid_WAV = []
#         self.valid_G_LMN = []
#         self.valid_ORDER = []
#         self.valid_GRATING = []
#         self.valid_DISTANCE = []
#         self.valid_OP = []
#         self.valid_TOP_S = []
#         self.valid_TOP = []
#         self.valid_ALPHA = []
#         self.valid_BULK_TRANS = []
#         self.valid_RP = []
#         self.valid_RS = []
#         self.valid_TP = []
#         self.valid_TS = []
#         self.valid_TTBE = []
#         self.valid_TT = []
#         self.invalid_RayWave = []
#         self.invalid_CCC = pv.MultiBlock()
#         self.invalid_SURFACE = []
#         self.invalid_NAME = []
#         self.invalid_GLASS = []
#         self.invalid_S_XYZ = []
#         self.invalid_T_XYZ = []
#         self.invalid_XYZ = []
#         self.invalid_OST_XYZ = []
#         self.invalid_S_LMN = []
#         self.invalid_LMN = []
#         self.invalid_R_LMN = []
#         self.invalid_N0 = []
#         self.invalid_N1 = []
#         self.invalid_WAV = []
#         self.invalid_G_LMN = []
#         self.invalid_ORDER = []
#         self.invalid_GRATING = []
#         self.invalid_DISTANCE = []
#         self.invalid_OP = []
#         self.invalid_TOP_S = []
#         self.invalid_TOP = []
#         self.invalid_ALPHA = []
#         self.invalid_BULK_TRANS = []
#         self.invalid_RP = []
#         self.invalid_RS = []
#         self.invalid_TP = []
#         self.invalid_TS = []
#         self.invalid_TTBE = []
#         self.invalid_TT = []



#     def pick(self, N_ELEMENT=(- 1)):
#         """pick.

#         Parameters
#         ----------
#         N_ELEMENT :
#             N_ELEMENT
#         """
#         self.numsup = (self.nelements - 1)
#         self.xyz = self.valid_XYZ
#         self.lmn = self.valid_LMN
#         self.s = self.valid_SURFACE
#         if ((N_ELEMENT < 0) or (N_ELEMENT > self.numsup)):
#             N_ELEMENT = self.numsup
#         else:
#             N_ELEMENT = N_ELEMENT
#         AA = []
#         BB = []
#         for k in self.s:
#             aa = np.argwhere((k == N_ELEMENT))
#             aa = np.squeeze(aa)
#             AA.append(aa)
#             BB.append(np.size(aa))
#         AA = np.asarray(AA)
#         BB = np.asarray(BB)
#         if (N_ELEMENT != 0):
#             BB = np.argwhere((BB == 1))
#         else:
#             BB = np.argwhere((BB == 0))
#         X = []
#         Y = []
#         Z = []
#         L = []
#         M = []
#         N = []
#         for c in BB:
#             for d in c:
#                 ray0 = self.xyz[d]
#                 [x1, y1, z1] = ray0[N_ELEMENT]
#                 X.append(x1)
#                 Y.append(y1)
#                 Z.append(z1)
#                 ray1 = self.lmn[d]
#                 if (N_ELEMENT != 0):
#                     el = (N_ELEMENT - 1)
#                 else:
#                     el = 0
#                 [l1, m1, n1] = ray1[el]
#                 L.append(l1)
#                 M.append(m1)
#                 N.append(n1)
#         return (np.asarray(X), np.asarray(Y), np.asarray(Z), np.asarray(L), np.asarray(M), np.asarray(N))






# class raykeeper_draw():
#     """raykeeper.
#     """


#     def __init__(self, System):
#         """__init__.

#         Parameters
#         ----------
#         System :
#             System
#         """

#         self.SYSTEM = System
#         self.clean()


#     def valid(self):
#         """valid.
#         """
#         z = np.argwhere((self.vld == 1))
#         return z

#     def push(self):
#         """push.
#         """

#         self.RayWave.append(self.SYSTEM.Wave)
#         self.CC.append(self.SYSTEM.ray_SurfHits)



#     def clean(self):
#         """clean.
#         """

#         self.RayWave = []
#         self.CC = []

